"""PV system module."""

# clean

# Generic/Built-in
import datetime
import enum
import math
import os
from dataclasses import dataclass
from typing import Any, List, Optional

import numpy as np
import pandas as pd
import pvlib
from dataclasses_json import dataclass_json

# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import log
from hisim import utils
from hisim.component import ConfigBase, OpexCostDataClass, CapexCostDataClass
from hisim.components.weather import Weather
from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiTagEnumClass, KpiEntry

__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"

"""
The functions cited in this module are at some degree based on the tsib project:

[tsib-kotzur]:
Kotzur, Leander, Detlef Stolten, and Hermann-Josef Wagner.
Future grid load of the residential building sector. No. RWTH-2018-231872. Lehrstuhl für Brennstoffzellen (FZ Jülich), 2019.
ID: http://hdl.handle.net/2128/21115
    http://nbn-resolving.org/resolver?verb=redirect&identifier=urn:nbn:de:0001-2019020614

The implementation of the tsib project can be found under the following repository:
https://github.com/FZJ-IEK3-VSA/tsib
"""


class PVLibModuleAndInverterEnum(enum.Enum):
    """Class to determine what pvlib database for phtotovoltaic modules and inverters should be used.

    https://pvlib-python.readthedocs.io/en/v0.9.0/generated/pvlib.pvsystem.retrieve_sam.html.
    """

    SANDIA_MODULE_DATABASE = 1
    SANDIA_INVERTER_DATABASE = 2
    CEC_MODULE_DATABASE = 3
    CEC_INVERTER_DATABASE = 4
    ANTON_DRIESSE_INVERTER_DATABASE = 5


@dataclass_json
@dataclass
class PVSystemConfig(ConfigBase):
    """PVSystemConfig class."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return PVSystem.get_full_classname()

    building_name: str
    name: str
    time: int
    location: str
    module_name: str
    integrate_inverter: bool
    inverter_name: str
    module_database: PVLibModuleAndInverterEnum
    inverter_database: PVLibModuleAndInverterEnum
    power_in_watt: float
    azimuth: float
    tilt: float
    # [0..1], how much pv potential is used
    share_of_maximum_pv_potential: float
    load_module_data: bool
    source_weight: int
    #: CO2 footprint of investment in kg
    co2_footprint: float
    #: cost for investment in Euro
    cost: float
    # maintenance cost as share of investment [0..1]
    maintenance_cost_as_percentage_of_investment: float
    #: lifetime in years
    lifetime: float
    predictive: bool
    predictive_control: bool
    prediction_horizon: Optional[int]

    @classmethod
    def get_default_pv_system(
        cls,
        name: str = "PVSystem",
        power_in_watt: float = 10e3,
        share_of_maximum_pv_potential: float = 1.0,
        location: str = "Aachen",
        building_name: str = "BUI1",
    ) -> "PVSystemConfig":
        """Gets a default PV system."""
        power_in_watt = power_in_watt * share_of_maximum_pv_potential
        return PVSystemConfig(
            building_name=building_name,
            time=2019,
            power_in_watt=power_in_watt,
            load_module_data=False,
            integrate_inverter=True,
            module_database=PVLibModuleAndInverterEnum.SANDIA_MODULE_DATABASE,
            inverter_database=PVLibModuleAndInverterEnum.SANDIA_INVERTER_DATABASE,
            module_name="Hanwha HSL60P6-PA-4-250T [2013]",
            inverter_name="ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_",
            name=name,
            azimuth=180,
            tilt=30,
            share_of_maximum_pv_potential=share_of_maximum_pv_potential,
            source_weight=0,
            location=location,
            co2_footprint=power_in_watt * 1e-3 * 330.51,  # value from emission_factros_and_costs_devices.csv
            cost=power_in_watt * 1e-3 * 794.41,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.01,  # source: https://solarenergie.de/stromspeicher/preise
            lifetime=25,  # value from emission_factros_and_costs_devices.csv
            predictive=False,
            predictive_control=False,
            prediction_horizon=None,
        )

    @classmethod
    def get_scaled_pv_system(
        cls,
        rooftop_area_in_m2: float,
        name: str = "PVSystem",
        share_of_maximum_pv_potential: float = 1.0,
        module_name: str = "Hanwha HSL60P6-PA-4-250T [2013]",
        module_database: PVLibModuleAndInverterEnum = PVLibModuleAndInverterEnum.SANDIA_MODULE_DATABASE,
        location: str = "Aachen",
        building_name: str = "BUI1",
        load_module_data: bool = False,
    ) -> "PVSystemConfig":
        """Gets a default PV system with scaling according to rooftop area."""
        total_pv_power_in_watt = cls.size_pv_system(
            rooftop_area_in_m2=rooftop_area_in_m2,
            share_of_maximum_pv_potential=share_of_maximum_pv_potential,
            module_name=module_name,
            module_database=module_database,
        )
        return PVSystemConfig(
            building_name=building_name,
            time=2019,
            power_in_watt=total_pv_power_in_watt,
            load_module_data=load_module_data,
            module_name=module_name,
            integrate_inverter=True,
            inverter_name="ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_",
            module_database=module_database,
            inverter_database=PVLibModuleAndInverterEnum.SANDIA_INVERTER_DATABASE,
            name=name,
            azimuth=180,
            tilt=30,
            share_of_maximum_pv_potential=share_of_maximum_pv_potential,
            source_weight=0,
            location=location,
            co2_footprint=total_pv_power_in_watt * 1e-3 * 330.51,  # value from emission_factros_and_costs_devices.csv
            cost=total_pv_power_in_watt * 1e-3 * 794.41,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.01,  # source: https://solarenergie.de/stromspeicher/preise
            lifetime=25,  # value from emission_factros_and_costs_devices.csv
            predictive=False,
            predictive_control=False,
            prediction_horizon=None,
        )

    @classmethod
    def size_pv_system(
        cls,
        rooftop_area_in_m2: float,
        share_of_maximum_pv_potential: float,
        module_name: str,
        module_database: PVLibModuleAndInverterEnum,
    ) -> float:
        """Size the pv system according to the rooftop type and the share of the maximum pv power that should be used."""

        # get area and power of module
        if (
            module_name == "Hanwha HSL60P6-PA-4-250T [2013]"
            and module_database == PVLibModuleAndInverterEnum.SANDIA_MODULE_DATABASE
        ):
            module_area_in_m2 = 1.65
            module_power_in_watt = 250
            # this is equal to an efficiency of 15,15%

        elif (
            module_name == "Trina Solar TSM-410DE09"
            and module_database == PVLibModuleAndInverterEnum.CEC_MODULE_DATABASE
        ):
            module_area_in_m2 = 1.91
            module_power_in_watt = 410
            # this is equal to an efficiency of 21,47%

        # pv module efficiency calculation see:
        # https://www.ess-kempfle.de/ratgeber/ertrag/pv-ertrag/#:~:text=So%20berechnen%20Sie%20den%20Wirkungsgrad,liegt%20bei%201.000%20W%2Fm%C2%B2.
        else:
            raise ValueError(
                f"Module name or module database {module_name} {module_database} not given in this function. Please check or add your module information."
            )

        # scale rooftop area with limiting factor due to shading and obstacles like chimneys etc.
        # see p.18 in following paper https://www.mdpi.com/1996-1073/15/15/5536 (Stanleys work)
        limiting_factor_for_rooftop = 0.6
        effective_rooftop_area_in_m2 = rooftop_area_in_m2 * limiting_factor_for_rooftop

        total_pv_power_in_watt = (
            effective_rooftop_area_in_m2 / module_area_in_m2 * module_power_in_watt
        ) * share_of_maximum_pv_potential

        return total_pv_power_in_watt


class PVSystem(cp.Component):
    """Simulates PV Output based on weather data and peak power.

    Parameters
    ----------
    time : int, optional
        Simulation timeline. The default is 2019.
    location : str, optional
        Object Location with temperature and solar data. The default is "Aachen".
    power : float, optional
        Power in kWp to be provided by the PV System. The default is 10E3.
    load_module_data : bool
        Access the PV data base (True) or not (False). The default is False
    module_name : str, optional
        The default is "Hanwha_HSL60P6_PA_4_250T__2013_"
    integrate_inverter, bool, optional
        Consider inverter efficiency in the calculation (True) or not (False). The default is True.
    inverter_name : str, optional
        The default is "ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_".
    azimuth : float, optional
        Panel azimuth from north in °. The default is 180°.
    tilt : float, optional
        Panel tilt from horizontal. The default is 90°.
    source_weight : int, optional
        Weight of component, relevant if there is more than one PV System, defines hierachy in control. The default is 1.
    name : str, optional
        Name of pv panel within simulation. The default is 'PVSystem'

    """

    # Inputs
    TemperatureOutside = "TemperatureOutside"
    DirectNormalIrradiance = "DirectNormalIrradiance"
    DirectNormalIrradianceExtra = "DirectNormalIrradianceExtra"
    DiffuseHorizontalIrradiance = "DiffuseHorizontalIrradiance"
    GlobalHorizontalIrradiance = "GlobalHorizontalIrradiance"
    Azimuth = "Azimuth"
    ApparentZenith = "ApparentZenith"
    WindSpeed = "WindSpeed"

    # Outputs
    ElectricityOutput = "ElectricityOutput"
    # Additional output channels must not contain 'ElectricityOutput' or dynamic components will fail.
    ElectricitityEnergyOutput = "ElectricitityEnergyOutput"

    # Similar components to connect to:
    # 1. Weather
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: PVSystemConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(display_in_webtool=True),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.pvconfig = config
        self.ac_power_ratios_for_all_timesteps_data: List = []
        self.ac_power_ratios_for_all_timesteps_output: List = []
        self.cache_filepath: str
        self.modules: Any
        self.inverter: Any
        self.inverters: Any
        self.module: Any
        self.coordinates: Any
        self.data_length: int = self.my_simulation_parameters.timesteps
        self.temperature_model_parameters = pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS["sapm"][
            "open_rack_glass_glass"
        ]
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.t_out_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureOutside,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            True,
        )

        self.dni_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.DirectNormalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )

        self.dni_extra_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.DirectNormalIrradianceExtra,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )

        self.dhi_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.DiffuseHorizontalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )

        self.ghi_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.GlobalHorizontalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )

        self.azimuth_channel: cp.ComponentInput = self.add_input(
            self.component_name, self.Azimuth, lt.LoadTypes.ANY, lt.Units.DEGREES, True
        )

        self.apparent_zenith_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ApparentZenith,
            lt.LoadTypes.ANY,
            lt.Units.DEGREES,
            True,
        )

        self.wind_speed_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WindSpeed,
            lt.LoadTypes.SPEED,
            lt.Units.METER_PER_SECOND,
            True,
        )

        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=[
                lt.InandOutputType.ELECTRICITY_PRODUCTION,
                lt.ComponentType.PV,
            ],
            output_description=f"here a description for PV {self.ElectricityOutput} will follow.",
        )

        self.electricity_energy_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricitityEnergyOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            postprocessing_flag=[
                lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
            ],
            output_description=f"Here a description for PV {self.ElectricitityEnergyOutput} will follow.",
        )

        self.add_default_connections(self.get_default_connections_from_weather())

    @staticmethod
    def get_default_config(
        power_in_watt: float = 10e3,
        source_weight: int = 1,
        share_of_maximum_pv_potential: float = 1.0,
        building_name: str = "BUI1",
    ) -> Any:
        """Get default config."""
        config = PVSystemConfig(
            building_name=building_name,
            name="PVSystem",
            time=2019,
            location="Aachen",
            module_name="Hanwha HSL60P6-PA-4-250T [2013]",
            integrate_inverter=True,
            inverter_name="ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_",
            module_database=PVLibModuleAndInverterEnum.SANDIA_MODULE_DATABASE,
            inverter_database=PVLibModuleAndInverterEnum.SANDIA_INVERTER_DATABASE,
            power_in_watt=power_in_watt,
            azimuth=180,
            tilt=30,
            share_of_maximum_pv_potential=share_of_maximum_pv_potential,
            load_module_data=False,
            source_weight=source_weight,
            co2_footprint=power_in_watt * 1e-3 * 130.7,  # value from emission_factros_and_costs_devices.csv
            cost=power_in_watt * 1e-3 * 535.81,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.01,  # source: https://solarenergie.de/stromspeicher/preise
            lifetime=25,  # value from emission_factros_and_costs_devices.csv
            prediction_horizon=None,
            predictive=False,
            predictive_control=False,
        )
        return config

    @staticmethod
    def get_cost_capex(config: PVSystemConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:
        """Returns investment cost, CO2 emissions and lifetime."""
        seconds_per_year = 365 * 24 * 60 * 60
        capex_per_simulated_period = (config.cost / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )
        device_co2_footprint_per_simulated_period = (config.co2_footprint / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )

        capex_cost_data_class = CapexCostDataClass(
            capex_investment_cost_in_euro=config.cost,
            device_co2_footprint_in_kg=config.co2_footprint,
            lifetime_in_years=config.lifetime,
            capex_investment_cost_for_simulated_period_in_euro=capex_per_simulated_period,
            device_co2_footprint_for_simulated_period_in_kg=device_co2_footprint_per_simulated_period,
            kpi_tag=KpiTagEnumClass.ROOFTOP_PV
        )
        return capex_cost_data_class

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        # pylint: disable=unused-argument
        """Calculate OPEX costs, consisting of maintenance costs for PV."""
        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=self.calc_maintenance_cost(),
            co2_footprint_in_kg=0,
            consumption_in_kwh=0,
            loadtype=lt.LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.ROOFTOP_PV
        )

        return opex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []

    def get_default_connections_from_weather(self):
        """Get default connections from weather."""

        connections = []
        weather_classname = Weather.get_classname()
        connections.append(
            cp.ComponentConnection(
                PVSystem.TemperatureOutside,
                weather_classname,
                Weather.TemperatureOutside,
            )
        )
        connections.append(
            cp.ComponentConnection(
                PVSystem.DirectNormalIrradiance,
                weather_classname,
                Weather.DirectNormalIrradiance,
            )
        )
        connections.append(
            cp.ComponentConnection(
                PVSystem.DirectNormalIrradianceExtra,
                weather_classname,
                Weather.DirectNormalIrradianceExtra,
            )
        )
        connections.append(
            cp.ComponentConnection(
                PVSystem.DiffuseHorizontalIrradiance,
                weather_classname,
                Weather.DiffuseHorizontalIrradiance,
            )
        )
        connections.append(
            cp.ComponentConnection(
                PVSystem.GlobalHorizontalIrradiance,
                weather_classname,
                Weather.GlobalHorizontalIrradiance,
            )
        )
        connections.append(cp.ComponentConnection(PVSystem.Azimuth, weather_classname, Weather.Azimuth))
        connections.append(cp.ComponentConnection(PVSystem.ApparentZenith, weather_classname, Weather.ApparentZenith))
        connections.append(cp.ComponentConnection(PVSystem.WindSpeed, weather_classname, Weather.WindSpeed))
        return connections

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""

        # check if results could be found in cache and if the lists have the right length
        if (
            hasattr(self, "ac_power_ratios_for_all_timesteps_output")
            and len(self.ac_power_ratios_for_all_timesteps_output) == self.data_length
        ):
            stsv.set_output_value(
                self.electricity_output_channel,
                self.ac_power_ratios_for_all_timesteps_output[timestep] * self.pvconfig.power_in_watt,
            )
            stsv.set_output_value(
                self.electricity_energy_output_channel,
                self.ac_power_ratios_for_all_timesteps_output[timestep]
                * self.pvconfig.power_in_watt
                * self.my_simulation_parameters.seconds_per_timestep
                / 3600,
            )

        # calculate pv system outputs with pvlib
        else:
            dni = stsv.get_input_value(self.dni_channel)
            dni_extra = stsv.get_input_value(self.dni_extra_channel)
            dhi = stsv.get_input_value(self.dhi_channel)
            ghi = stsv.get_input_value(self.ghi_channel)
            azimuth = stsv.get_input_value(self.azimuth_channel)
            temperature = stsv.get_input_value(self.t_out_channel)
            wind_speed = stsv.get_input_value(self.wind_speed_channel)
            apparent_zenith = stsv.get_input_value(self.apparent_zenith_channel)

            ac_power_ratio = self.simphotovoltaic_two(
                dni_extra=dni_extra,
                dni=dni,
                dhi=dhi,
                ghi=ghi,
                azimuth=azimuth,
                apparent_zenith=apparent_zenith,
                temperature=temperature,
                wind_speed=wind_speed,
                surface_azimuth=self.pvconfig.azimuth,
                surface_tilt=self.pvconfig.tilt,
            )

            ac_power_in_watt = ac_power_ratio * self.pvconfig.power_in_watt

            # if you wanted to access the temperature forecast from the weather component:
            # val = self.simulation_repository.get_entry(Weather.Weather_Temperature_Forecast_24h)

            stsv.set_output_value(self.electricity_output_channel, ac_power_in_watt)
            stsv.set_output_value(
                self.electricity_energy_output_channel,
                ac_power_in_watt * self.my_simulation_parameters.seconds_per_timestep / 3600,
            )

            # cache results at the end of the simulation
            self.ac_power_ratios_for_all_timesteps_data[timestep] = ac_power_ratio

            if timestep + 1 == self.data_length:
                dict_with_results = {
                    "output_power": self.ac_power_ratios_for_all_timesteps_data,
                }

                database = pd.DataFrame(
                    dict_with_results,
                    columns=[
                        "output_power",
                    ],
                )

                database.to_csv(self.cache_filepath, sep=",", decimal=".", index=False)

        if self.pvconfig.predictive_control and self.pvconfig.prediction_horizon is not None:
            last_forecast_timestep = int(
                timestep + self.pvconfig.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep
            )
            if last_forecast_timestep > len(self.ac_power_ratios_for_all_timesteps_output):
                last_forecast_timestep = len(self.ac_power_ratios_for_all_timesteps_output)
            pvforecast = [
                self.ac_power_ratios_for_all_timesteps_output[t] * self.pvconfig.power_in_watt
                for t in range(timestep, last_forecast_timestep)
            ]
            self.simulation_repository.set_dynamic_entry(
                component_type=lt.ComponentType.PV,
                source_weight=self.pvconfig.source_weight,
                entry=pvforecast,
            )

            if timestep == 1:
                # delete weather data for PV preprocessing from dictionary -> save memory
                if SingletonSimRepository().exist_entry(
                    key=SingletonDictKeyEnum.WEATHERDIRECTNORMALIRRADIANCEEXTRAYEARLYFORECAST
                ):
                    SingletonSimRepository().delete_entry(
                        key=SingletonDictKeyEnum.WEATHERDIRECTNORMALIRRADIANCEEXTRAYEARLYFORECAST
                    )
                    SingletonSimRepository().delete_entry(
                        key=SingletonDictKeyEnum.WEATHERDIRECTNORMALIRRADIANCEYEARLYFORECAST
                    )
                    SingletonSimRepository().delete_entry(
                        key=SingletonDictKeyEnum.WEATHERDIFFUSEHORIZONTALIRRADIANCEYEARLYFORECAST
                    )
                    SingletonSimRepository().delete_entry(
                        key=SingletonDictKeyEnum.WEATHERGLOBALHORIZONTALIRRADIANCEYEARLYFORECAST
                    )
                    SingletonSimRepository().delete_entry(key=SingletonDictKeyEnum.WEATHERAZIMUTHYEARLYFORECAST)
                    SingletonSimRepository().delete_entry(key=SingletonDictKeyEnum.WEATHERAPPARENTZENITHYEARLYFORECAST)
                    SingletonSimRepository().delete_entry(
                        key=SingletonDictKeyEnum.WEATHERTEMPERATUREOUTSIDEYEARLYFORECAST
                    )
                    SingletonSimRepository().delete_entry(key=SingletonDictKeyEnum.WEATHERWINDSPEEDYEARLYFORECAST)

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def write_to_report(self):
        """Write to the report."""
        return self.pvconfig.get_string_dict()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the component for the simulation."""
        file_exists, self.cache_filepath = utils.get_cache_file(
            "PVSystem", self.pvconfig, self.my_simulation_parameters
        )

        if file_exists:
            log.information("Get PV results from cache.")
            self.ac_power_ratios_for_all_timesteps_output = pd.read_csv(self.cache_filepath, sep=",", decimal=".")[
                "output_power"
            ].tolist()

            if len(self.ac_power_ratios_for_all_timesteps_output) != self.my_simulation_parameters.timesteps:
                raise Exception(
                    "Reading the cached PV values seems to have failed. Expected "
                    + str(self.my_simulation_parameters.timesteps)
                    + " values, but got "
                    + str(len(self.ac_power_ratios_for_all_timesteps_output))
                )
        else:
            if SingletonSimRepository().exist_entry(key=SingletonDictKeyEnum.LOCATION):
                SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.LOCATION)
            else:
                raise KeyError(
                    "The key weather_location was not found in the singleton sim repository."
                    "Please check in your system setup if the weather component was added to the simulator before the pv system."
                )

            # read module from pvlib database online or read from csv files in hisim/inputs/photovoltaic/data_processed
            self.module = self.get_modules_from_database(
                module_database=self.pvconfig.module_database,
                load_module_data=self.pvconfig.load_module_data,
                module_name=self.pvconfig.module_name,
            )

            # read inverter from pvlib database online or read from csv files in hisim/inputs/photovoltaic/data_processed
            self.inverter = self.get_inverters_from_database(
                inverter_database=self.pvconfig.inverter_database,
                load_module_data=self.pvconfig.load_module_data,
                inverter_name=self.pvconfig.inverter_name,
            )

            # when predictive control is activated, the PV simulation is run beforhand to make forecasting easier
            if self.pvconfig.predictive_control:
                # get yearly weather data from dictionary
                dni_extra = SingletonSimRepository().get_entry(
                    key=SingletonDictKeyEnum.WEATHERDIRECTNORMALIRRADIANCEEXTRAYEARLYFORECAST
                )
                dni = SingletonSimRepository().get_entry(
                    key=SingletonDictKeyEnum.WEATHERDIRECTNORMALIRRADIANCEYEARLYFORECAST
                )
                dhi = SingletonSimRepository().get_entry(
                    key=SingletonDictKeyEnum.WEATHERDIFFUSEHORIZONTALIRRADIANCEYEARLYFORECAST
                )
                ghi = SingletonSimRepository().get_entry(
                    key=SingletonDictKeyEnum.WEATHERGLOBALHORIZONTALIRRADIANCEYEARLYFORECAST
                )
                azimuth = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.WEATHERAZIMUTHYEARLYFORECAST)
                apparent_zenith = SingletonSimRepository().get_entry(
                    key=SingletonDictKeyEnum.WEATHERAPPARENTZENITHYEARLYFORECAST
                )
                temperature = SingletonSimRepository().get_entry(
                    key=SingletonDictKeyEnum.WEATHERTEMPERATUREOUTSIDEYEARLYFORECAST
                )
                wind_speed = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.WEATHERWINDSPEEDYEARLYFORECAST)

                x_simplephotovoltaic = []
                for i in range(self.my_simulation_parameters.timesteps):
                    # calculate outputs
                    ac_power_ratio = self.simphotovoltaic_two(
                        dni_extra=dni_extra[i],
                        dni=dni[i],
                        dhi=dhi[i],
                        ghi=ghi[i],
                        azimuth=azimuth[i],
                        apparent_zenith=apparent_zenith[i],
                        temperature=temperature[i],
                        wind_speed=wind_speed[i],
                        surface_azimuth=self.pvconfig.azimuth,
                        surface_tilt=self.pvconfig.tilt,
                    )

                    # append lists
                    x_simplephotovoltaic.append(ac_power_ratio)

                self.ac_power_ratios_for_all_timesteps_output = x_simplephotovoltaic

                # cache predictive control results
                dict_with_results = {
                    "output_power": self.ac_power_ratios_for_all_timesteps_output,
                }

                database = pd.DataFrame(
                    dict_with_results,
                    columns=[
                        "output_power",
                    ],
                )

                database.to_csv(self.cache_filepath, sep=",", decimal=".", index=False)

            else:
                # create empty result lists as a preparation for caching in i_simulate

                self.ac_power_ratios_for_all_timesteps_data = [0] * self.my_simulation_parameters.timesteps

        if self.pvconfig.predictive:
            pv_forecast_yearly = [
                self.ac_power_ratios_for_all_timesteps_output[t] * self.pvconfig.power_in_watt
                for t in range(self.my_simulation_parameters.timesteps)
            ]
            SingletonSimRepository().set_entry(key=SingletonDictKeyEnum.PVFORECASTYEARLY, entry=pv_forecast_yearly)

    def interpolate(self, pd_database: Any, year: Any) -> Any:
        """Interpolates."""
        lastday = pd.Series(
            pd_database[-1],
            index=[pd.to_datetime(datetime.datetime(year, 12, 31, 22, 59), utc=True).tz_convert("Europe/Berlin")],
        )

        pd_database = pd_database.append(lastday)
        pd_database = pd_database.sort_index()
        return pd_database.resample("1T").asfreq().interpolate(method="linear").tolist()

    def get_modules_from_database(self, module_database: Any, load_module_data: bool, module_name: str) -> Any:
        """Get modules from pvlib module database."""

        # get modules from pvlib database online (TODO: test if this works, it has not been fully tested yet)
        if load_module_data is True:
            if module_database == PVLibModuleAndInverterEnum.SANDIA_MODULE_DATABASE:
                modules = pvlib.pvsystem.retrieve_sam(name="SandiaMod")
            elif module_database == PVLibModuleAndInverterEnum.CEC_MODULE_DATABASE:
                modules = pvlib.pvsystem.retrieve_sam(name="CECMod")
            else:
                raise KeyError(f"The module database {module_database} is not integrated in the PV component here.")

            # choose module from modules database
            module = modules[module_name]

        # get modules from input data csv files
        else:
            if module_database == PVLibModuleAndInverterEnum.SANDIA_MODULE_DATABASE:
                modules = pd.read_csv(
                    os.path.join(utils.HISIMPATH["photovoltaic"]["sandia_modules_new"]),
                )

            # note that you can import cec module database but the calculations of the pv system can only be done with sandia!
            elif module_database == PVLibModuleAndInverterEnum.CEC_MODULE_DATABASE:
                modules = pd.read_csv(os.path.join(utils.HISIMPATH["photovoltaic"]["cec_modules"]))
                log.warning(
                    "You can import pv cec modules but the pvlib calculations (simphotovoltaic_two) only work with the sandia modules unfortunately."
                )
            else:
                raise KeyError(f"The module database {module_database} is not integrated in the PV component here.")

            # choose module from modules database
            module = modules.loc[modules["Name"] == module_name]

            # transform column object types to numeric types
            for column in module.columns:
                module.loc[:, column] = pd.to_numeric(module.loc[:, column], errors="coerce")

            # transform module dataframe to dict
            module = module.to_dict(orient="records")[0]

        return module

    def get_inverters_from_database(self, inverter_database: Any, load_module_data: bool, inverter_name: str) -> Any:
        """Get inverters from pvlib module database."""

        # get inverters from pvlib database online
        if load_module_data is True:
            if inverter_database in (
                PVLibModuleAndInverterEnum.SANDIA_INVERTER_DATABASE,
                PVLibModuleAndInverterEnum.CEC_INVERTER_DATABASE,
            ):
                # get inverter data (for both sandia and cec inverters the same database is taken):
                # see docs: https://pvlib-python.readthedocs.io/en/v0.9.0/generated/pvlib.pvsystem.retrieve_sam.html
                inverters = pvlib.pvsystem.retrieve_sam("CECInverter")
                inverter = inverters[inverter_name]
            elif inverter_database == PVLibModuleAndInverterEnum.ANTON_DRIESSE_INVERTER_DATABASE:
                inverters = pvlib.pvsystem.retrieve_sam("ADRInverter")
                inverter = inverters[inverter_name]
            else:
                raise KeyError(f"The inverter database {inverter_database} is not integrated in the PV component here.")

        # get inverters from input data csv files
        else:
            # this is the old csv file used in hisim
            if inverter_database == PVLibModuleAndInverterEnum.SANDIA_INVERTER_DATABASE:
                inverters = pd.read_csv(
                    os.path.join(utils.HISIMPATH["photovoltaic"]["sandia_inverters"]),
                    index_col=0,
                )
                # choose inverter from inverters database
                inverter = inverters[inverter_name]
                # transform to numeric types
                inverter = pd.to_numeric(inverter, errors="coerce")

            # this would be the new one, but not tested yet
            elif inverter_database == PVLibModuleAndInverterEnum.CEC_INVERTER_DATABASE:
                inverters = pd.read_csv(
                    os.path.join(utils.HISIMPATH["photovoltaic"]["cec_inverters"]),
                )
                # choose inverter from inverters database
                inverter = inverters.loc[inverters["Name"] == inverter_name]

                # transform column object types to numeric types
                for column in inverter.columns:
                    inverter[column] = pd.to_numeric(inverter[column], errors="coerce")

                # transform inverter dataframe to dict
                inverter = inverter.to_dict(orient="records")[0]

            else:
                raise KeyError(f"The inverter database {inverter_database} is not integrated in the PV component here.")

        return inverter

    def simphotovoltaic_two(
        self,
        dni_extra=None,
        dni=None,
        dhi=None,
        ghi=None,
        azimuth=None,
        apparent_zenith=None,
        temperature=None,
        wind_speed=None,
        surface_tilt=30,
        surface_azimuth=180,
        albedo=0.2,
    ):
        r"""Simulates a defined PV array with the Sandia PV Array Performance Model.

        The implementation is done in accordance with following tutorial:
        https://github.com/pvlib/pvlib-python/blob/master/docs/tutorials/tmy_to_power.ipynb
        https://pvlib-python.readthedocs.io/en/stable/reference/generated/pvlib.pvsystem.sapm.html#pvlib.pvsystem.sapm

        Based on the tsib project @[tsib-kotzur] (Check header)

        Parameters
        ----------
        tmy_data: pandas.DataFrame(), required
            Weatherfile in the format of a tmy file.
        surface_tilt: int or float, optional (default:30)
            Tilt angle of of the array in degree.
        surface_azimuth: int or float, optional (default:180)
            Azimuth angle of of the array in degree. 180 degree means south,
            90 degree east and 270 west.
        albedo: float, optional (default: 0.2)
            Reflection coefficient of the surrounding area.
        losses: float, optional (default: 0.1)
            Losses due to soiling, mismatch, diode connections, dc wiring etc.
        load_module_data: Boolean, optional (default: False)
            If True the module data base is loaded from the Sandia Website.
            Otherwise it is loaded from this relative path
                '\\profiles\\PV-Modules\\sandia_modules.csv'.
        module_name: str, optional (default:'Hanwha_HSL60P6_PA_4_250T__2013_')
            Module name. The string must be existens in Sandia Module database.
        integrateInverter: bool, optional (default: True)
            If an inverter shall be added to the simulation, providing the photovoltaic output after the inverter.
        inverter_name: str, optional (default: 'ABB__MICRO_0_25_I_OUTD_US_208_208V__CEC_2014_')
            Type of inverter.
        apparent_zenith: Any
            Apparent zenith.
        azimuth: int, float
            Azimuth.
        dni: Any
            direct normal irradiance.
        ghi: Any
            global horizontal irradiance.
        dhi: Any
            direct horizontal irradiance.
        dni_extra: Any
            direct normal irradiance extra.
        temperature: Any
            tempertaure.
        wind_speed: Any
            wind_speed.

        Returns
        -------
        ac_power: Any
            ac power

        """
        # automatic pd time series in future pvlib version
        # calculate airmass
        airmass = pvlib.atmosphere.get_relative_airmass(apparent_zenith)
        # use perez model to calculate the plane of array diffuse sky radiation
        poa_sky_diffuse = pvlib.irradiance.perez(
            surface_tilt,
            surface_azimuth,
            dhi,
            np.float64(dni),
            dni_extra,
            apparent_zenith,
            azimuth,
            airmass,
        )
        # calculate ground diffuse with specified albedo
        poa_ground_diffuse = pvlib.irradiance.get_ground_diffuse(surface_tilt, ghi, albedo=albedo)
        # calculate angle of incidence
        aoi = pvlib.irradiance.aoi(surface_tilt, surface_azimuth, apparent_zenith, azimuth)
        # calculate plane of array irradiance
        poa_irrad = pvlib.irradiance.poa_components(aoi, np.float64(dni), poa_sky_diffuse, poa_ground_diffuse)
        # calculate pv cell and module temperature

        pvtemps = pvlib.temperature.sapm_cell(
            poa_irrad["poa_global"],
            temperature,
            wind_speed,
            **self.temperature_model_parameters,
        )

        # calculate effective irradiance on pv module
        sapm_irr = pvlib.pvsystem.sapm_effective_irradiance(
            module=self.module,
            poa_direct=poa_irrad["poa_direct"],
            poa_diffuse=poa_irrad["poa_diffuse"],
            airmass_absolute=airmass,
            aoi=aoi,
        )
        # calculate pv performance
        sapm_out = pvlib.pvsystem.sapm(
            sapm_irr,
            module=self.module,
            temp_cell=pvtemps,
        )
        # calculate peak load of single module [W]
        module_peak_load_in_watt = self.module["Impo"] * self.module["Vmpo"]
        ac_power_ratio: float

        if self.pvconfig.integrate_inverter:
            # calculate load after inverter
            inverter_load_in_watt = pvlib.inverter.sandia(
                inverter=self.inverter, v_dc=sapm_out["v_mp"], p_dc=sapm_out["p_mp"]
            )
            # if inverter load is nan, make it zero otherwise ac_power_ratio will be nan also
            if math.isnan(inverter_load_in_watt):
                inverter_load_in_watt = 0

            ac_power_ratio = inverter_load_in_watt / module_peak_load_in_watt
        else:
            # load in [kW/kWp]
            ac_power_ratio = sapm_out["p_mp"] / module_peak_load_in_watt

        if math.isnan(ac_power_ratio):  # type: ignore
            ac_power_ratio = 0.0

        return ac_power_ratio



""" rSOC controller. """

# clean
import os
import json
from typing import Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from hisim.component import ConfigBase, Component, ComponentInput, ComponentOutput, SingleTimeStepValues, DisplayConfig
from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters


# from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum


__authors__ = "Franz Oldopp"
__copyright__ = "Copyright 2023, IEK-3"
__credits__ = ["Franz Oldopp"]
__license__ = "MIT"
__version__ = "0.5"
__maintainer__ = "Franz Oldopp"
__email__ = "f.oldopp@fz-juelich.de"
__status__ = "development"


@dataclass_json
@dataclass
class RsocControllerConfig(ConfigBase):
    """Config of the rSOC Controller."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return RsocController.get_full_classname()

    building_name: str
    name: str

    nom_load_soec: float
    min_load_soec: float
    max_load_soec: float
    warm_start_time_soec: float
    cold_start_time_soec: float
    switching_time_from_soec_to_sofc: float

    nom_power_sofc: float
    min_power_sofc: float
    max_power_sofc: float
    warm_start_time_sofc: float
    cold_start_time_sofc: float
    switching_time_from_sofc_to_soec: float

    @staticmethod
    def read_config(rsoc_name):
        """Opens the according JSON-file, based on the rSOC_name."""

        config_file = os.path.join(utils.HISIMPATH["inputs"], "rSOC_manufacturer_config.json")
        with open(config_file, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
            return data.get("rSOC variants", {}).get(rsoc_name, {})

    @classmethod
    def config_rsoc(
        cls,
        rsoc_name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Initializes the config variables based on the JSON-file."""

        config_json = cls.read_config(rsoc_name)
        config = RsocControllerConfig(
            building_name=building_name,
            name="rSCO l1 Controller",
            nom_load_soec=config_json.get("nom_load_soec", 0.0),
            min_load_soec=config_json.get("min_load_soec", 0.0),
            max_load_soec=config_json.get("max_load_soec", 0.0),
            warm_start_time_soec=config_json.get("warm_start_time_soec", 0.0),
            cold_start_time_soec=config_json.get("cold_start_time_soec", 0.0),
            switching_time_from_soec_to_sofc=config_json.get("switching_time_from_soec_to_sofc", 0.0),
            nom_power_sofc=config_json.get("nom_power_sofc", 0.0),
            min_power_sofc=config_json.get("min_power_sofc", 0.0),
            max_power_sofc=config_json.get("max_power_sofc", 0.0),
            warm_start_time_sofc=config_json.get("warm_start_time_sofc", 0.0),
            cold_start_time_sofc=config_json.get("cold_start_time_sofc", 0.0),
            switching_time_from_sofc_to_soec=config_json.get("switching_time_from_sofc_to_soec", 0.0),
        )
        return config


class RsocController(Component):
    """rSOC Controller class."""

    # Inputs
    ProvidedPower = "ProvidedPower"
    PowerDemand = "PowerDemand"

    # Outputs
    PowerToSOEC = "PowerToSOEC"
    StateToRSOC = "StateToRSOC"
    DemandToSOFC = "DemandToSOFC"
    StateToSOFC = "StateToSOFC"
    PowerVsDemand = "PowerVsDemand"
    CurtailedLoad = "CurtailedLoad"
    CurtailedPower = "CurtailedPower"
    TotalOffCount = "# of times the system was switched off"
    TotalStandbyCount = "# of times the system was to standby mode"
    SwitchCount = "SwitchCount"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: RsocControllerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initialize class."""
        self.rsoccontrollerconfig = config

        self.name = config.name

        self.nom_load_soec = config.nom_load_soec
        self.min_load_soec = config.min_load_soec
        self.max_load_soec = config.max_load_soec
        self.warm_start_time_soec = config.warm_start_time_soec
        self.cold_start_time_soec = config.cold_start_time_soec
        self.switching_time_from_soec_to_sofc = config.switching_time_from_soec_to_sofc

        self.nom_power_sofc = config.nom_power_sofc
        self.min_power_sofc = config.min_power_sofc
        self.max_power_sofc = config.max_power_sofc
        self.warm_start_time_sofc = config.warm_start_time_sofc
        self.cold_start_time_sofc = config.cold_start_time_sofc
        self.switching_time_from_sofc_to_soec = config.switching_time_from_sofc_to_soec

        self.cold_start_time = self.cold_start_time_soec
        self.warm_start_time = self.warm_start_time_soec
        self.standby_load = 100.0

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        # =================================================================================================================================
        # Input channels

        # Getting the load input
        self.provided_power: ComponentInput = self.add_input(
            self.component_name,
            RsocController.ProvidedPower,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            True,
        )
        # Getting the demand input
        self.power_demand: ComponentInput = self.add_input(
            self.component_name,
            RsocController.PowerDemand,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            False,
        )

        # =================================================================================================================================
        # Output channels

        self.state_rsoc: ComponentOutput = self.add_output(
            self.component_name,
            RsocController.StateToRSOC,
            lt.LoadTypes.ACTIVATION,
            lt.Units.ANY,
            output_description="State to the RSOC",
        )

        self.current_delta: ComponentOutput = self.add_output(
            self.component_name,
            RsocController.PowerVsDemand,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            output_description="Current delta between provided power and power demand",
        )

        self.curtailed_load: ComponentOutput = self.add_output(
            self.component_name,
            RsocController.CurtailedLoad,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            output_description="Curtailed load",
        )
        self.curtailed_power: ComponentOutput = self.add_output(
            self.component_name,
            RsocController.CurtailedPower,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            output_description="Curtailed power",
        )
        self.total_switch_count: ComponentOutput = self.add_output(
            self.component_name,
            RsocController.SwitchCount,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Total count of time spend switching between operation modes",
        )
        self.total_off_count: ComponentOutput = self.add_output(
            self.component_name,
            RsocController.TotalOffCount,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Total count of off cycles",
        )
        self.total_standby_count: ComponentOutput = self.add_output(
            self.component_name,
            RsocController.TotalStandbyCount,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Total count of standby cycles",
        )

        # =================================================================================================================================
        # Initialize variables

        self.standby_count = 0.0
        self.current_state = "OFF"  # standby
        self.warm_count = 0.0
        self.cold_count = 0.0
        self.initial_state = 0
        self.curtailed_load_count = 0.0
        self.cold_up_count = 0.0
        self.warm_up_count = 0.0
        self.standby_count = 0.0
        self.off_count = 0.0
        self.activation_runtime = 0.0
        self.elapsed_time = 0.0
        self.switch_start_time = 0.0
        self.rsoc_state = "SOFC"
        self.total_switch_time_count = 0.0

        self.standby_count_previous = self.standby_count
        self.current_state_previous = self.current_state
        self.warm_count_previous = self.warm_count
        self.cold_count_previous = self.cold_count
        self.initial_state_previous = self.initial_state
        self.curtailed_load_count_previous = self.curtailed_load_count
        self.cold_up_count_previous = self.cold_up_count
        self.warm_up_count_previous = self.warm_up_count
        self.standby_count_previous = self.standby_count
        self.off_count_previous = self.off_count
        self.activation_runtime_previous = self.activation_runtime
        self.elapsed_time_previous = self.elapsed_time
        self.rsoc_state_previous = self.rsoc_state
        self.total_switch_time_count_previous = self.total_switch_time_count

    def i_prepare_simulation(self) -> None:
        """Prepare the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.standby_count_previous = self.standby_count
        self.current_state_previous = self.current_state
        self.warm_count_previous = self.warm_count
        self.cold_count_previous = self.cold_count
        self.initial_state_previous = self.initial_state
        self.curtailed_load_count_previous = self.curtailed_load_count
        self.cold_up_count_previous = self.cold_up_count
        self.warm_up_count_previous = self.warm_up_count
        self.standby_count_previous = self.standby_count
        self.off_count_previous = self.off_count
        self.activation_runtime_previous = self.activation_runtime
        self.elapsed_time_previous = self.elapsed_time
        self.rsoc_state_previous = self.rsoc_state
        self.total_switch_time_count_previous = self.total_switch_time_count

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.standby_count = self.standby_count_previous
        self.current_state = self.current_state_previous
        self.warm_count = self.warm_count_previous
        self.cold_count = self.cold_count_previous
        self.initial_state = self.initial_state_previous
        self.curtailed_load_count = self.curtailed_load_count_previous
        self.cold_up_count = self.cold_up_count_previous
        self.warm_up_count = self.warm_up_count_previous
        self.standby_count = self.standby_count_previous
        self.off_count = self.off_count_previous
        self.activation_runtime = self.activation_runtime_previous
        self.elapsed_time = self.elapsed_time_previous
        self.rsoc_state = self.rsoc_state_previous
        self.total_switch_time_count = self.total_switch_time_count_previous

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""
        if force_convergence:
            return
        """
        self.nom_load_soec = config.nom_load_soec
        self.min_load_soec = config.min_load_soec
        self.max_load_soec = config.max_load_soec
        self.warm_start_time_soec = config.warm_start_time_soec
        self.cold_start_time_soec = config.cold_start_time_soec
        self.switching_time_from_soec_to_sofc = config.switching_time_from_soec_to_sofc

        self.nom_power_sofc = config.nom_power_sofc
        self.min_power_sofc = config.min_power_sofc
        self.max_power_sofc = config.max_power_sofc
        self.warm_start_time_sofc = config.warm_start_time_sofc
        self.cold_start_time_sofc = config.cold_start_time_sofc
        self.switching_time_from_sofc_to_soec = config.switching_time_from_sofc_to_soec

        """
        self.cold_start_time = self.cold_start_time_soec
        self.warm_start_time = self.warm_start_time_soec
        self.standby_load = 100.0

        if timestep < 5:
            print(self.min_load_soec)
            print(self.min_power_sofc)

        power = stsv.get_input_value(self.provided_power)
        # if -self.min_load_soec < power < 0.0:
        #    print("should switch OFF SOEC")
        # elif 0.0 < power < self.min_power_sofc:
        #    print("should switch OFF SOFC")

        if power < -self.min_load_soec or self.min_power_sofc < power:
            if self.current_state == "OFF":
                self.activation_runtime = self.cold_start_time
                self.current_state = "Starting"
                self.off_count += 1
            elif self.current_state == "STANDBY":
                self.activation_runtime = self.warm_start_time
                self.current_state = "Starting"
                self.standby_count += 1
            elif self.current_state == "Starting":
                if self.activation_runtime <= self.my_simulation_parameters.seconds_per_timestep:
                    self.activation_runtime = 0.0
                    self.current_state = "ON"
                else:
                    self.activation_runtime -= self.my_simulation_parameters.seconds_per_timestep
            else:
                self.activation_runtime = 0.0
                self.current_state = "ON"
        elif power in (-self.min_load_soec, self.min_power_sofc):
            if self.current_state == "ON":
                self.current_state = "STANDBY"
            else:
                self.current_state = self.current_state
        else:
            self.current_state = "OFF"

        """
        if power != 0.0:
            if self.current_state == "OFF":
                self.activation_runtime = self.cold_start_time
                self.current_state = "Starting"
            elif self.current_state == "STANDBY":
                self.activation_runtime = self.warm_start_time
                self.current_state = "Starting"
            elif self.current_state == "Starting":
                if self.activation_runtime <= self.my_simulation_parameters.seconds_per_timestep:
                    self.activation_runtime = 0.0
                    self.current_state = "ON"
                else:
                    self.activation_runtime -= self.my_simulation_parameters.seconds_per_timestep
            else:
                self.activation_runtime = 0.0
                self.current_state = "ON"
        """

        if power < 0.0 and self.current_state == "ON":
            if self.rsoc_state == "SOEC":
                power_to_rsco = power
                state_to_rsoc = 1
            elif self.rsoc_state in ("SOFC", "Switching to SOFC"):
                self.elapsed_time = self.switching_time_from_sofc_to_soec
                self.rsoc_state = "Switching to SOEC"
                power_to_rsco = 0.0
                state_to_rsoc = 0
                self.total_switch_time_count += self.switching_time_from_sofc_to_soec
            else:
                if self.elapsed_time <= self.my_simulation_parameters.seconds_per_timestep:
                    self.elapsed_time = 0.0
                    self.rsoc_state = "SOEC"
                    power_to_rsco = 0.0
                    state_to_rsoc = 0
                else:
                    self.elapsed_time -= self.my_simulation_parameters.seconds_per_timestep
                    power_to_rsco = 0.0
                    state_to_rsoc = 0

        elif power > 0.0 and self.current_state == "ON":
            if self.rsoc_state == "SOFC":
                power_to_rsco = power
                state_to_rsoc = 1
            elif self.rsoc_state in ("SOEC", "Switching to SOEC"):
                self.elapsed_time = self.switching_time_from_soec_to_sofc
                self.rsoc_state = "Switching to SOFC"
                power_to_rsco = 0.0
                state_to_rsoc = 0
                self.total_switch_time_count += self.switching_time_from_soec_to_sofc
            else:
                if self.elapsed_time <= self.my_simulation_parameters.seconds_per_timestep:
                    self.elapsed_time = 0.0
                    self.rsoc_state = "SOFC"
                    power_to_rsco = 0.0
                    state_to_rsoc = 0
                else:
                    self.elapsed_time -= self.my_simulation_parameters.seconds_per_timestep
                    power_to_rsco = 0.0
                    state_to_rsoc = 0
        elif power < 0.0 and self.current_state == "STANDBY":
            power_to_rsco = self.min_load_soec
            state_to_rsoc = 0
        elif power > 0.0 and self.current_state == "STANDBY":
            power_to_rsco = self.min_power_sofc
            state_to_rsoc = 0
        else:
            power_to_rsco = 0.0
            state_to_rsoc = -1

        # stsv.set_output_value(self.power_soec, power_to_soec)
        stsv.set_output_value(self.state_rsoc, state_to_rsoc)
        # stsv.set_output_value(self.demand_sofc, power_to_sofc)
        # stsv.set_output_value(self.state_sofc, state_to_sofc)
        stsv.set_output_value(self.current_delta, power_to_rsco)
        stsv.set_output_value(self.total_switch_count, self.total_switch_time_count)
        stsv.set_output_value(self.total_off_count, self.off_count)
        stsv.set_output_value(self.total_standby_count, self.standby_count)



#!/usr/bin/env python

"""The setup script."""
# clean
from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("requirements.txt", encoding="utf-8") as requirements_file:
    requirements = requirements_file.read().splitlines()

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]


setup(
    author="Noah Pflugradt",
    author_email="n.pflugradt@fz-juelich.de",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="ETHOS.HiSim is a house infrastructure simulator",
    entry_points={
        "console_scripts": [
            "hisim=hisim.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="hisim",
    name="hisim",
    packages=find_packages(include=["hisim", "hisim.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/FZJ-IEK3-VSA/HiSim",
    version="1.2.1",
    zip_safe=False,
)



""" Class for the simulation repository. """
# clean
from typing import Any, Dict

from hisim import loadtypes as lt


class SimRepository:

    """Class for exchanging information across all components."""

    def __init__(self) -> None:
        """Initializes the SimRepository."""
        self.my_dict: Dict[str, Any] = {}
        self.my_dynamic_dict: Dict[lt.ComponentType, Dict[int, Any]] = {elem: {} for elem in lt.ComponentType}

    def set_entry(self, key: str, entry: Any) -> None:
        """Sets an entry in the SimRepository."""
        self.my_dict[key] = entry

    def get_entry(self, key: str) -> Any:
        """Gets an entry from the SimRepository."""
        return self.my_dict[key]

    def exist_entry(self, key: str) -> bool:
        """Checks if an entry exists."""
        if key in self.my_dict:
            return True
        return False

    def delete_entry(self, key: str) -> None:
        """Deletes an existing entry."""
        self.my_dict.pop(key)

    def set_dynamic_entry(self, component_type: lt.ComponentType, source_weight: int, entry: Any) -> None:
        """Sets a dynamic entry."""
        self.my_dynamic_dict[component_type][source_weight] = entry

    def get_dynamic_entry(self, component_type: lt.ComponentType, source_weight: int) -> Any:
        """Gets a dynmaic entry."""
        component = self.my_dynamic_dict.get(component_type, None)
        if component is None:
            return None
        value = component.get(source_weight, None)
        return value

    def get_dynamic_component_weights(self, component_type: lt.ComponentType) -> list:
        """Gets weights for dynamic components."""
        return list(self.my_dynamic_dict[component_type].keys())

    def delete_dynamic_entry(self, component_type: lt.ComponentType, source_weight: int) -> Any:
        """Deletes a dynamic component entry."""
        self.my_dynamic_dict[component_type].pop(source_weight)

    def clear(self):
        """Clears all dictionaries at the end of the simulation to enable garbage collection and reduce memory consumption."""
        self.my_dict.clear()
        del self.my_dict
        self.my_dynamic_dict.clear()
        del self.my_dynamic_dict



"""Advanced heat pump module.

See library on https://github.com/FZJ-IEK3-VSA/hplib/tree/main/hplib
"""

import hashlib

# clean
import importlib
from dataclasses import dataclass
from typing import Any, List, Optional, Dict

import pandas as pd
from dataclass_wizard import JSONWizard
from dataclasses_json import dataclass_json
from hplib import hplib as hpl

# Import modules from HiSim
from hisim.component import (
    Component,
    ComponentInput,
    ComponentOutput,
    SingleTimeStepValues,
    ConfigBase,
    ComponentConnection,
    OpexCostDataClass,
    CapexCostDataClass,
    DisplayConfig,
)
from hisim.components import weather, simple_hot_water_storage, heat_distribution_system
from hisim.components.heat_distribution_system import HeatDistributionSystemType
from hisim.loadtypes import LoadTypes, Units, InandOutputType, OutputPostprocessingRules
from hisim.units import (
    Quantity,
    Watt,
    Celsius,
    Seconds,
    Kilogram,
    Euro,
    Years,
)

from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiHelperClass, KpiTagEnumClass

__authors__ = "Tjarko Tjaden, Hauke Hoops, Kai Rösken"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = "..."
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Tjarko Tjaden"
__email__ = "tjarko.tjaden@hs-emden-leer.de"
__status__ = "development"


@dataclass_json
@dataclass
class HeatPumpHplibConfig(ConfigBase):
    """HeatPumpHPLibConfig."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return HeatPumpHplib.get_full_classname()

    building_name: str
    name: str
    model: str
    group_id: int
    heating_reference_temperature_in_celsius: Quantity[float, Celsius]  # before t_in
    flow_temperature_in_celsius: Quantity[float, Celsius]  # before t_out_val
    set_thermal_output_power_in_watt: Quantity[float, Watt]  # before p_th_set
    cycling_mode: bool
    minimum_running_time_in_seconds: Optional[Quantity[int, Seconds]]
    minimum_idle_time_in_seconds: Optional[Quantity[int, Seconds]]
    #: CO2 footprint of investment in kg
    co2_footprint: Quantity[float, Kilogram]
    #: cost for investment in Euro
    cost: Quantity[float, Euro]
    #: lifetime in years
    lifetime: Quantity[float, Years]
    # maintenance cost as share of investment [0..1]
    maintenance_cost_as_percentage_of_investment: float

    @classmethod
    def get_default_generic_advanced_hp_lib(
        cls,
        set_thermal_output_power_in_watt: Quantity[float, Watt] = Quantity(8000, Watt),
        heating_reference_temperature_in_celsius: Quantity[float, Celsius] = Quantity(-7.0, Celsius),
        building_name: str = "BUI1",
    ) -> "HeatPumpHplibConfig":
        """Gets a default HPLib Heat Pump.

        see default values for air/water hp on:
        https://github.com/FZJ-IEK3-VSA/hplib/blob/main/hplib/hplib.py l.135 "fit_p_th_ref.
        """
        return HeatPumpHplibConfig(
            building_name=building_name,
            name="AdvancedHeatPumpHPLib",
            model="Generic",
            group_id=1,
            heating_reference_temperature_in_celsius=heating_reference_temperature_in_celsius,
            flow_temperature_in_celsius=Quantity(52, Celsius),
            set_thermal_output_power_in_watt=set_thermal_output_power_in_watt,
            cycling_mode=True,
            minimum_running_time_in_seconds=Quantity(3600, Seconds),
            minimum_idle_time_in_seconds=Quantity(3600, Seconds),
            # value from emission_factors_and_costs_devices.csv
            co2_footprint=Quantity(set_thermal_output_power_in_watt.value * 1e-3 * 165.84, Kilogram),
            # value from emission_factors_and_costs_devices.csv
            cost=Quantity(set_thermal_output_power_in_watt.value * 1e-3 * 1513.74, Euro),
            lifetime=Quantity(10, Years),  # value from emission_factors_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.025,  # source:  VDI2067-1
        )

    @classmethod
    def get_scaled_advanced_hp_lib(
        cls,
        heating_load_of_building_in_watt: Quantity[float, Watt],
        heating_reference_temperature_in_celsius: Quantity[float, Celsius] = Quantity(-7.0, Celsius),
        name: str = "AdvancedHeatPumpHPLib",
        building_name: str = "BUI1",
    ) -> "HeatPumpHplibConfig":
        """Gets a default heat pump with scaling according to heating load of the building_name."""

        set_thermal_output_power_in_watt: Quantity[float, Watt] = heating_load_of_building_in_watt

        return HeatPumpHplibConfig(
            building_name=building_name,
            name=name,
            model="Generic",
            group_id=1,
            heating_reference_temperature_in_celsius=heating_reference_temperature_in_celsius,
            flow_temperature_in_celsius=Quantity(52, Celsius),
            set_thermal_output_power_in_watt=set_thermal_output_power_in_watt,
            cycling_mode=True,
            minimum_running_time_in_seconds=Quantity(3600, Seconds),
            minimum_idle_time_in_seconds=Quantity(3600, Seconds),
            # value from emission_factros_and_costs_devices.csv
            co2_footprint=Quantity(set_thermal_output_power_in_watt.value * 1e-3 * 165.84, Kilogram),
            # value from emission_factros_and_costs_devices.csv
            cost=Quantity(set_thermal_output_power_in_watt.value * 1e-3 * 1513.74, Euro),
            # value from emission_factros_and_costs_devices.csv
            lifetime=Quantity(10, Years),
            maintenance_cost_as_percentage_of_investment=0.025,  # source:  VDI2067-1
        )


class HeatPumpHplib(Component):
    """Simulate the heat pump.

    Outputs are heat pump efficiency (cop) as well as electrical (p_el) and
    thermal power (p_th), massflow (m_dot) and output temperature (t_out).
    Relevant simulation parameters are loaded within the init for a
    specific or generic heat pump type.
    """

    # Inputs
    OnOffSwitch = "OnOffSwitch"  # 1 = on, 0 = 0ff
    TemperatureInputPrimary = "TemperatureInputPrimary"  # °C
    TemperatureInputSecondary = "TemperatureInputSecondary"  # °C
    TemperatureAmbient = "TemperatureAmbient"  # °C

    # Outputs
    ThermalOutputPower = "ThermalOutputPower"  # W
    ThermalOutputEnergy = "ThermalOutputEnergy"  # Wh
    ElectricalInputPower = "ElectricalInputPower"  # W
    ElectricalInputPowerForHeating = "ElectricalInputPowerForHeating"  # W
    ElectricalInputPowerForCooling = "ElectricalInputPowerForCooling"  # W
    ElectricalInputEnergy = "ElectricalInputEnergy"  # Wh
    COP = "COP"  # -
    EER = "EER"  # -
    TemperatureOutput = "TemperatureOutput"  # °C
    TemperatureInputWarmWater = "TemperatureInputWarmWater"  # °C
    MassFlowOutput = "MassFlowOutput"  # kg/s
    TimeOnHeating = "TimeOnHeating"  # s
    TimeOnCooling = "TimeOnCooling"  # s
    TimeOff = "TimeOff"  # s

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: HeatPumpHplibConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Loads the parameters of the specified heat pump.

        model : str
            Name of the heat pump model or "Generic".
        group_id : numeric, default 0
            only for model "Generic": Group ID for subtype of heat pump. [1-6].
        t_in : numeric, default 0
            only for model "Generic": Input temperature :math:`T` at primary side of the heat pump. [°C]
        t_out_val : numeric, default 0
            only for model "Generic": Output temperature :math:`T` at secondary side of the heat pump. [°C]
        p_th_set : numeric, default 0
            only for model "Generic": Thermal output power at setpoint t_in, t_out. [W]

        """

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        # caching for hplib simulation
        self.calculation_cache: Dict = {}

        self.model = config.model

        self.group_id = config.group_id

        self.t_in = int(config.heating_reference_temperature_in_celsius.value)

        self.t_out_val = int(config.flow_temperature_in_celsius.value)

        self.p_th_set = int(config.set_thermal_output_power_in_watt.value)

        self.cycling_mode = config.cycling_mode

        self.minimum_running_time_in_seconds = (
            config.minimum_running_time_in_seconds.value
            if config.minimum_running_time_in_seconds
            else config.minimum_running_time_in_seconds
        )

        self.minimum_idle_time_in_seconds = (
            config.minimum_idle_time_in_seconds.value
            if config.minimum_idle_time_in_seconds
            else config.minimum_idle_time_in_seconds
        )

        # Component has states
        self.state = HeatPumpState(time_on_heating=0, time_off=0, time_on_cooling=0, on_off_previous=0)
        self.previous_state = self.state.self_copy()

        # Load parameters from heat pump database
        self.parameters = hpl.get_parameters(self.model, self.group_id, self.t_in, self.t_out_val, self.p_th_set)

        # Define component inputs
        self.on_off_switch: ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.OnOffSwitch,
            load_type=LoadTypes.ANY,
            unit=Units.ANY,
            mandatory=True,
        )

        self.t_in_primary: ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.TemperatureInputPrimary,
            load_type=LoadTypes.TEMPERATURE,
            unit=Units.CELSIUS,
            mandatory=True,
        )

        self.t_in_secondary: ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.TemperatureInputSecondary,
            load_type=LoadTypes.TEMPERATURE,
            unit=Units.CELSIUS,
            mandatory=True,
        )

        self.t_amb: ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.TemperatureAmbient,
            load_type=LoadTypes.TEMPERATURE,
            unit=Units.CELSIUS,
            mandatory=True,
        )

        # Define component outputs
        self.p_th: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ThermalOutputPower,
            load_type=LoadTypes.HEATING,
            unit=Units.WATT,
            output_description="Thermal output power in Watt",
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )

        self.q_th: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ThermalOutputEnergy,
            load_type=LoadTypes.HEATING,
            unit=Units.WATT_HOUR,
            output_description=("Thermal output enery in Watthours"),
        )

        self.p_el: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricalInputPower,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            postprocessing_flag=[
                InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED,
                OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
            ],
            output_description="Electricity input power in Watt",
        )
        self.p_el_heating: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricalInputPowerForHeating,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
            output_description="Electricity input power for heating in Watt",
        )
        self.p_el_cooling: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricalInputPowerForCooling,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
            output_description="Electricity input power for cooling in Watt",
        )

        self.e_el: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricalInputEnergy,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT_HOUR,
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
            output_description="Electricity input energy in Watthours",
        )

        self.cop: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.COP,
            load_type=LoadTypes.ANY,
            unit=Units.ANY,
            output_description="COP",
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )
        self.eer: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.EER,
            load_type=LoadTypes.ANY,
            unit=Units.ANY,
            output_description="EER",
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )
        self.t_out: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TemperatureOutput,
            load_type=LoadTypes.HEATING,
            unit=Units.CELSIUS,
            output_description="Temperature Output in °C",
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )
        self.t_in_warm_water: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TemperatureInputWarmWater,
            load_type=LoadTypes.HEATING,
            unit=Units.CELSIUS,
            output_description="Temperature Input in °C",
            postprocessing_flag=[
                OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
            ],
        )

        self.m_dot: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.MassFlowOutput,
            load_type=LoadTypes.VOLUME,
            unit=Units.KG_PER_SEC,
            output_description="Mass flow output",
        )

        self.time_on_heating: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TimeOnHeating,
            load_type=LoadTypes.TIME,
            unit=Units.SECONDS,
            output_description="Time turned on for heating",
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )
        self.time_on_cooling: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TimeOnCooling,
            load_type=LoadTypes.TIME,
            unit=Units.SECONDS,
            output_description="Time turned on for cooling",
            postprocessing_flag=[OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )

        self.time_off: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TimeOff,
            load_type=LoadTypes.TIME,
            unit=Units.SECONDS,
            output_description="Time turned off",
        )

        self.add_default_connections(self.get_default_connections_from_heat_pump_controller())
        self.add_default_connections(self.get_default_connections_from_weather())
        self.add_default_connections(self.get_default_connections_from_simple_hot_water_storage())

    def get_default_connections_from_heat_pump_controller(self,):
        """Get default connections."""

        connections = []
        hpc_classname = HeatPumpHplibController.get_classname()
        connections.append(
            ComponentConnection(HeatPumpHplib.OnOffSwitch, hpc_classname, HeatPumpHplibController.State,)
        )
        return connections

    def get_default_connections_from_weather(self,):
        """Get default connections."""

        connections = []
        weather_classname = weather.Weather.get_classname()
        connections.append(
            ComponentConnection(
                HeatPumpHplib.TemperatureAmbient, weather_classname, weather.Weather.DailyAverageOutsideTemperatures,
            )
        )

        connections.append(
            ComponentConnection(
                HeatPumpHplib.TemperatureInputPrimary,
                weather_classname,
                weather.Weather.DailyAverageOutsideTemperatures,
            )
        )
        return connections

    def get_default_connections_from_simple_hot_water_storage(self,):
        """Get simple hot water storage default connections."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.simple_hot_water_storage"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "SimpleHotWaterStorage")
        connections = []
        hws_classname = component_class.get_classname()
        connections.append(
            ComponentConnection(
                HeatPumpHplib.TemperatureInputSecondary, hws_classname, component_class.WaterTemperatureToHeatGenerator,
            )
        )
        return connections

    def write_to_report(self):
        """Write configuration to the report."""
        return self.config.get_string_dict()

    def i_save_state(self) -> None:
        """Save state."""
        self.previous_state = self.state.self_copy()
        # pass

    def i_restore_state(self) -> None:
        """Restore state."""
        self.state = self.previous_state.self_copy()
        # pass

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doubelcheck."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepare simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""

        # Load input values
        on_off: float = stsv.get_input_value(self.on_off_switch)
        t_in_primary = stsv.get_input_value(self.t_in_primary)
        t_in_secondary = stsv.get_input_value(self.t_in_secondary)
        t_amb = stsv.get_input_value(self.t_amb)
        time_on_heating = self.state.time_on_heating
        time_on_cooling = self.state.time_on_cooling
        time_off = self.state.time_off

        # cycling means periodic turning on and off of the heat pump
        if self.cycling_mode is True:
            # Parameter
            time_on_min = self.minimum_running_time_in_seconds  # [s]
            time_off_min = self.minimum_idle_time_in_seconds
            on_off_previous = self.state.on_off_previous

            if time_on_min is None or time_off_min is None:
                raise ValueError(
                    """When the cycling mode is true, the minimum running time and minimum idle time of the heat pump
                    must be given an integer value."""
                )

            # Overwrite on_off to realize minimum time of or time off
            if on_off_previous == 1 and time_on_heating < time_on_min:
                on_off = 1
            elif on_off_previous == -1 and time_on_cooling < time_on_min:
                on_off = -1
            elif on_off_previous == 0 and time_off < time_off_min:
                on_off = 0

        # heat pump is turned on and off only according to heat pump controller
        elif self.cycling_mode is False:
            pass
        else:
            raise ValueError("Cycling mode of the advanced hplib unknown.")

        # OnOffSwitch
        if on_off == 1:
            results = self.get_cached_results_or_run_hplib_simulation(
                t_in_primary=t_in_primary,
                t_in_secondary=t_in_secondary,
                parameters=self.parameters,
                t_amb=t_amb,
                mode=1,
            )

            # Get outputs for heating mode
            p_th = results["P_th"].values[0]
            q_th = p_th * self.my_simulation_parameters.seconds_per_timestep / 3600
            p_el = results["P_el"].values[0]
            p_el_heating = p_el
            p_el_cooling = 0
            e_el = p_el * self.my_simulation_parameters.seconds_per_timestep / 3600
            cop = results["COP"].values[0]
            eer = results["EER"].values[0]
            t_out = results["T_out"].values[0]
            m_dot = results["m_dot"].values[0]
            time_on_heating = time_on_heating + self.my_simulation_parameters.seconds_per_timestep
            time_on_cooling = 0
            time_off = 0

        elif on_off == -1:
            # Calulate outputs for cooling mode
            results = self.get_cached_results_or_run_hplib_simulation(
                t_in_primary=t_in_primary,
                t_in_secondary=t_in_secondary,
                parameters=self.parameters,
                t_amb=t_amb,
                mode=2,
            )

            p_th = results["P_th"].values[0]
            q_th = p_th * self.my_simulation_parameters.seconds_per_timestep / 3600
            p_el = results["P_el"].values[0]
            p_el_heating = 0
            p_el_cooling = p_el
            e_el = p_el * self.my_simulation_parameters.seconds_per_timestep / 3600
            cop = results["COP"].values[0]
            eer = results["EER"].values[0]
            t_out = results["T_out"].values[0]
            m_dot = results["m_dot"].values[0]
            time_on_cooling = time_on_cooling + self.my_simulation_parameters.seconds_per_timestep
            time_on_heating = 0
            time_off = 0
        elif on_off == 0:
            # Calulate outputs for off mode
            p_th = 0
            q_th = 0
            p_el = 0
            p_el_heating = 0
            p_el_cooling = 0
            e_el = 0
            # None values or nans will cause troubles in post processing, that is why there are not used here
            # cop = None
            # t_out = None
            cop = 0
            eer = 0
            t_out = t_in_secondary
            m_dot = 0
            time_off = time_off + self.my_simulation_parameters.seconds_per_timestep
            time_on_heating = 0
            time_on_cooling = 0

        else:
            raise ValueError("Unknown mode for Advanced HPLib On_Off.")

        # write values for output time series
        stsv.set_output_value(self.p_th, p_th)
        stsv.set_output_value(self.q_th, q_th)
        stsv.set_output_value(self.p_el, p_el)
        stsv.set_output_value(self.p_el_heating, p_el_heating)
        stsv.set_output_value(self.p_el_cooling, p_el_cooling)
        stsv.set_output_value(self.e_el, e_el)
        stsv.set_output_value(self.cop, cop)
        stsv.set_output_value(self.eer, eer)
        stsv.set_output_value(self.t_out, t_out)
        stsv.set_output_value(self.t_in_warm_water, t_in_secondary)
        stsv.set_output_value(self.m_dot, m_dot)
        stsv.set_output_value(self.time_on_heating, time_on_heating)
        stsv.set_output_value(self.time_on_cooling, time_on_cooling)
        stsv.set_output_value(self.time_off, time_off)

        # write values to state
        self.state.time_on_heating = time_on_heating
        self.state.time_on_cooling = time_on_cooling
        self.state.time_off = time_off
        self.state.on_off_previous = on_off

    @staticmethod
    def get_cost_capex(config: HeatPumpHplibConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:
        """Returns investment cost, CO2 emissions and lifetime."""
        seconds_per_year = 365 * 24 * 60 * 60
        capex_per_simulated_period = (config.cost.value / config.lifetime.value) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )
        device_co2_footprint_per_simulated_period = (config.co2_footprint.value / config.lifetime.value) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )
        capex_cost_data_class = CapexCostDataClass(
            capex_investment_cost_in_euro=config.cost.value,
            device_co2_footprint_in_kg=config.co2_footprint.value,
            lifetime_in_years=config.lifetime.value,
            capex_investment_cost_for_simulated_period_in_euro=capex_per_simulated_period,
            device_co2_footprint_for_simulated_period_in_kg=device_co2_footprint_per_simulated_period,
            kpi_tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING
        )
        return capex_cost_data_class

    def get_cost_opex(self, all_outputs: List, postprocessing_results: pd.DataFrame,) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of maintenance costs.

        No electricity costs for components except for Electricity Meter,
        because part of electricity consumption is feed by PV
        """
        for index, output in enumerate(all_outputs):
            if (
                output.component_name == self.component_name
                and output.load_type == LoadTypes.ELECTRICITY
                and output.field_name == self.ElectricalInputPower
            ):  # Todo: check component name from system_setups: find another way of using only heatpump-outputs
                self.config.consumption_in_kwh = round(
                    sum(postprocessing_results.iloc[:, index])
                    * self.my_simulation_parameters.seconds_per_timestep
                    / 3.6e6,
                    1,
                )
        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=self.calc_maintenance_cost(),
            co2_footprint_in_kg=0,
            consumption_in_kwh=self.config.consumption_in_kwh,
            loadtype=LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING
        )

        return opex_cost_data_class

    def get_cached_results_or_run_hplib_simulation(
        self, t_in_primary: float, t_in_secondary: float, parameters: pd.DataFrame, t_amb: float, mode: int,
    ) -> Any:
        """Use caching of results of hplib simulation."""

        # rounding of variable values
        t_in_primary = round(t_in_primary, 1)
        t_in_secondary = round(t_in_secondary, 1)
        t_amb = round(t_amb, 1)

        my_data_class = CalculationRequest(
            t_in_primary=t_in_primary, t_in_secondary=t_in_secondary, t_amb=t_amb, mode=mode,
        )
        my_json_key = my_data_class.get_key()
        my_hash_key = hashlib.sha256(my_json_key.encode("utf-8")).hexdigest()

        if my_hash_key in self.calculation_cache:
            results = self.calculation_cache[my_hash_key]

        else:
            results = hpl.simulate(t_in_primary, t_in_secondary, parameters, t_amb, mode=mode)

            self.calculation_cache[my_hash_key] = results

        return results

    def get_component_kpi_entries(self, all_outputs: List, postprocessing_results: pd.DataFrame,) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""

        output_heating_energy_in_kilowatt_hour: float = 0.0
        output_cooling_energy_in_kilowatt_hour: float = 0.0
        electrical_energy_for_heating_in_kilowatt_hour: float = 1.0
        electrical_energy_for_cooling_in_kilowatt_hour: float = 1.0
        number_of_heat_pump_cycles: Optional[float] = None
        seasonal_performance_factor: Optional[float] = None
        seasonal_energy_efficiency_ratio: Optional[float] = None
        total_electrical_energy_input_in_kilowatt_hour: Optional[float] = None
        heating_time_in_hours: Optional[float] = None
        cooling_time_in_hours: Optional[float] = None

        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                number_of_heat_pump_cycles = self.get_heatpump_cycles(
                    output=output, index=index, postprocessing_results=postprocessing_results
                )
                if output.field_name == self.ThermalOutputPower and output.load_type == LoadTypes.HEATING:
                    # take only output values for heating
                    heating_output_power_values_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    # get energy from power
                    output_heating_energy_in_kilowatt_hour = KpiHelperClass.compute_total_energy_from_power_timeseries(
                        power_timeseries_in_watt=heating_output_power_values_in_watt,
                        timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                    )

                    # take only output values for cooling
                    cooling_output_power_values_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] < 0.0
                    ]
                    # for cooling enery use absolute value, not negative value
                    output_cooling_energy_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=cooling_output_power_values_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )

                elif output.field_name == self.ElectricalInputPowerForHeating:
                    # get electrical energie values for heating
                    electrical_energy_for_heating_in_kilowatt_hour = KpiHelperClass.compute_total_energy_from_power_timeseries(
                        power_timeseries_in_watt=postprocessing_results.iloc[:, index],
                        timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                    )

                elif output.field_name == self.ElectricalInputPowerForCooling:
                    # get electrical energie values for cooling
                    electrical_energy_for_cooling_in_kilowatt_hour = KpiHelperClass.compute_total_energy_from_power_timeseries(
                        power_timeseries_in_watt=postprocessing_results.iloc[:, index],
                        timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                    )

                elif output.field_name == self.TimeOnHeating:
                    heating_time_in_seconds = sum(postprocessing_results.iloc[:, index])
                    heating_time_in_hours = heating_time_in_seconds / 3600

                elif output.field_name == self.TimeOnCooling:
                    cooling_time_in_seconds = sum(postprocessing_results.iloc[:, index])
                    cooling_time_in_hours = cooling_time_in_seconds / 3600

        # calculate SPF
        if electrical_energy_for_heating_in_kilowatt_hour != 0.0:
            seasonal_performance_factor = (
                output_heating_energy_in_kilowatt_hour / electrical_energy_for_heating_in_kilowatt_hour
            )

        # calculate SEER
        if electrical_energy_for_cooling_in_kilowatt_hour != 0.0:
            seasonal_energy_efficiency_ratio = (
                output_cooling_energy_in_kilowatt_hour / electrical_energy_for_cooling_in_kilowatt_hour
            )

        # calculate total electricty input energy
        total_electrical_energy_input_in_kilowatt_hour = (
            electrical_energy_for_cooling_in_kilowatt_hour + electrical_energy_for_heating_in_kilowatt_hour
        )
        # make kpi entry
        number_of_heat_pump_cycles_entry = KpiEntry(
            name="Number of SH heat pump cycles",
            unit="-",
            value=number_of_heat_pump_cycles,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(number_of_heat_pump_cycles_entry)

        seasonal_performance_factor_entry = KpiEntry(
            name="Seasonal performance factor of SH heat pump",
            unit="-",
            value=seasonal_performance_factor,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(seasonal_performance_factor_entry)

        seasonal_energy_efficiency_entry = KpiEntry(
            name="Seasonal energy efficiency ratio of SH heat pump",
            unit="-",
            value=seasonal_energy_efficiency_ratio,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(seasonal_energy_efficiency_entry)

        heating_output_energy_heatpump_entry = KpiEntry(
            name="Heating output energy of SH heat pump",
            unit="kWh",
            value=output_heating_energy_in_kilowatt_hour,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(heating_output_energy_heatpump_entry)

        cooling_output_energy_heatpump_entry = KpiEntry(
            name="Cooling output energy of SH heat pump",
            unit="kWh",
            value=output_cooling_energy_in_kilowatt_hour,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(cooling_output_energy_heatpump_entry)

        electrical_input_energy_for_heating_entry = KpiEntry(
            name="Electrical input energy for heating of SH heat pump",
            unit="kWh",
            value=electrical_energy_for_heating_in_kilowatt_hour,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(electrical_input_energy_for_heating_entry)

        electrical_input_energy_for_cooling_entry = KpiEntry(
            name="Electrical input energy for cooling of SH heat pump",
            unit="kWh",
            value=electrical_energy_for_cooling_in_kilowatt_hour,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(electrical_input_energy_for_cooling_entry)

        electrical_input_energy_total_entry = KpiEntry(
            name="Total electrical input energy of SH heat pump",
            unit="kWh",
            value=total_electrical_energy_input_in_kilowatt_hour,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(electrical_input_energy_total_entry)

        heating_hours_entry = KpiEntry(
            name="Heating hours of SH heat pump",
            unit="h",
            value=heating_time_in_hours,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(heating_hours_entry)

        cooling_hours_entry = KpiEntry(
            name="Cooling hours of SH heat pump",
            unit="h",
            value=cooling_time_in_hours,
            tag=KpiTagEnumClass.HEATPUMP_SPACE_HEATING,
            description=self.component_name,
        )
        list_of_kpi_entries.append(cooling_hours_entry)

        return list_of_kpi_entries

    # make kpi entries and append to list
    def get_heatpump_cycles(self, output: Any, index: int, postprocessing_results: pd.DataFrame) -> float:
        """Get the number of cycles of the heat pump for the simulated period."""
        number_of_cycles = 0
        if output.field_name == self.TimeOff:
            for time_index, off_time in enumerate(postprocessing_results.iloc[:, index].values):
                try:
                    if off_time != 0 and postprocessing_results.iloc[:, index].values[time_index + 1] == 0:
                        number_of_cycles = number_of_cycles + 1

                except Exception:
                    pass

        return number_of_cycles


@dataclass
class HeatPumpState:
    """HeatPumpState class."""

    time_on_heating: int = 0
    time_off: int = 0
    time_on_cooling: int = 0
    on_off_previous: float = 0

    def self_copy(self,):
        """Copy the Heat Pump State."""
        return HeatPumpState(self.time_on_heating, self.time_off, self.time_on_cooling, self.on_off_previous)


# ===========================================================================
# try to implement a hplib controller l1
@dataclass_json
@dataclass
class HeatPumpHplibControllerL1Config(ConfigBase):
    """HeatPump Controller Config Class."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return HeatPumpHplibController.get_full_classname()

    building_name: str
    name: str
    mode: int
    set_heating_threshold_outside_temperature_in_celsius: Optional[float]
    set_cooling_threshold_outside_temperature_in_celsius: Optional[float]
    heat_distribution_system_type: Any

    @classmethod
    def get_default_generic_heat_pump_controller_config(
        cls,
        heat_distribution_system_type: Any,
        mode: int = 2,
        building_name: str = "BUI1",
        name: str = "HeatPumpController",
    ) -> "HeatPumpHplibControllerL1Config":
        """Gets a default Generic Heat Pump Controller."""
        return HeatPumpHplibControllerL1Config(
            building_name=building_name,
            name=name,
            mode=mode,
            set_heating_threshold_outside_temperature_in_celsius=16.0,
            set_cooling_threshold_outside_temperature_in_celsius=20.0,
            heat_distribution_system_type=heat_distribution_system_type,
        )


class HeatPumpHplibController(Component):
    """Heat Pump Controller.

    It takes data from other
    components and sends signal to the heat pump for
    activation or deactivation.
    On/off Switch with respect to water temperature from storage.
    """

    # Inputs
    WaterTemperatureInputFromHeatWaterStorage = "WaterTemperatureInputFromHeatWaterStorage"
    HeatingFlowTemperatureFromHeatDistributionSystem = "HeatingFlowTemperatureFromHeatDistributionSystem"

    DailyAverageOutsideTemperature = "DailyAverageOutsideTemperature"

    SimpleHotWaterStorageTemperatureModifier = "SimpleHotWaterStorageTemperatureModifier"

    # Outputs
    State = "State"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: HeatPumpHplibControllerL1Config,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Construct all the neccessary attributes."""
        self.heatpump_controller_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.heat_distribution_system_type = self.heatpump_controller_config.heat_distribution_system_type
        self.build(mode=self.heatpump_controller_config.mode,)

        self.water_temperature_input_channel: ComponentInput = self.add_input(
            self.component_name,
            self.WaterTemperatureInputFromHeatWaterStorage,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            True,
        )

        self.heating_flow_temperature_from_heat_distribution_system_channel: ComponentInput = self.add_input(
            self.component_name,
            self.HeatingFlowTemperatureFromHeatDistributionSystem,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            True,
        )
        self.daily_avg_outside_temperature_input_channel: ComponentInput = self.add_input(
            self.component_name, self.DailyAverageOutsideTemperature, LoadTypes.TEMPERATURE, Units.CELSIUS, True,
        )

        self.simple_hot_water_storage_temperature_modifier_channel: ComponentInput = self.add_input(
            self.component_name,
            self.SimpleHotWaterStorageTemperatureModifier,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            mandatory=False,
        )

        self.state_channel: ComponentOutput = self.add_output(
            self.component_name,
            self.State,
            LoadTypes.ANY,
            Units.ANY,
            output_description=f"here a description for {self.State} will follow.",
        )

        self.controller_heatpumpmode: Any
        self.previous_heatpump_mode: Any

        self.add_default_connections(self.get_default_connections_from_heat_distribution_controller())
        self.add_default_connections(self.get_default_connections_from_weather())
        self.add_default_connections(self.get_default_connections_from_simple_hot_water_storage())
        self.add_default_connections(self.get_default_connections_from_energy_management_system())

    def get_default_connections_from_heat_distribution_controller(self,):
        """Get default connections."""
        connections = []
        hdsc_classname = heat_distribution_system.HeatDistributionController.get_classname()
        connections.append(
            ComponentConnection(
                HeatPumpHplibController.HeatingFlowTemperatureFromHeatDistributionSystem,
                hdsc_classname,
                heat_distribution_system.HeatDistributionController.HeatingFlowTemperature,
            )
        )
        return connections

    def get_default_connections_from_weather(self,):
        """Get default connections."""
        connections = []
        weather_classname = weather.Weather.get_classname()
        connections.append(
            ComponentConnection(
                HeatPumpHplibController.DailyAverageOutsideTemperature,
                weather_classname,
                weather.Weather.DailyAverageOutsideTemperatures,
            )
        )
        return connections

    def get_default_connections_from_simple_hot_water_storage(self,):
        """Get simple hot water storage default connections."""
        connections = []
        hws_classname = simple_hot_water_storage.SimpleHotWaterStorage.get_classname()
        connections.append(
            ComponentConnection(
                HeatPumpHplibController.WaterTemperatureInputFromHeatWaterStorage,
                hws_classname,
                simple_hot_water_storage.SimpleHotWaterStorage.WaterTemperatureToHeatGenerator,
            )
        )
        return connections

    def get_default_connections_from_energy_management_system(self,):
        """Get energy management system default connections."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.controller_l2_energy_management_system"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "L2GenericEnergyManagementSystem")
        connections = []
        ems_classname = component_class.get_classname()
        connections.append(
            ComponentConnection(
                HeatPumpHplibController.SimpleHotWaterStorageTemperatureModifier,
                ems_classname,
                component_class.SpaceHeatingWaterStorageTemperatureModifier,
            )
        )
        return connections

    def build(self, mode: float,) -> None:
        """Build function.

        The function sets important constants and parameters for the calculations.
        """
        # Sth
        self.controller_heatpumpmode = "off"
        self.previous_heatpump_mode = self.controller_heatpumpmode

        # Configuration
        self.mode = mode

    def i_prepare_simulation(self) -> None:
        """Prepare the simulation."""
        pass

    def i_save_state(self) -> None:
        """Save the current state."""
        self.previous_heatpump_mode = self.controller_heatpumpmode

    def i_restore_state(self) -> None:
        """Restore the previous state."""
        self.controller_heatpumpmode = self.previous_heatpump_mode

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublecheck."""
        pass

    def write_to_report(self) -> List[str]:
        """Write important variables to report."""
        return self.heatpump_controller_config.get_string_dict()

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the heat pump comtroller."""

        if force_convergence:
            pass
        else:
            # Retrieves inputs

            water_temperature_input_from_heat_water_storage_in_celsius = stsv.get_input_value(
                self.water_temperature_input_channel
            )

            heating_flow_temperature_from_heat_distribution_system = stsv.get_input_value(
                self.heating_flow_temperature_from_heat_distribution_system_channel
            )

            daily_avg_outside_temperature_in_celsius = stsv.get_input_value(
                self.daily_avg_outside_temperature_input_channel
            )

            storage_temperature_modifier = stsv.get_input_value(
                self.simple_hot_water_storage_temperature_modifier_channel
            )

            # turning heat pump off when the average daily outside temperature is above a certain threshold (if threshold is set in the config)
            summer_heating_mode = self.summer_heating_condition(
                daily_average_outside_temperature_in_celsius=daily_avg_outside_temperature_in_celsius,
                set_heating_threshold_temperature_in_celsius=self.heatpump_controller_config.set_heating_threshold_outside_temperature_in_celsius,
            )

            # mode 1 is on/off controller
            if self.mode == 1:
                self.conditions_on_off(
                    water_temperature_input_in_celsius=water_temperature_input_from_heat_water_storage_in_celsius,
                    set_heating_flow_temperature_in_celsius=heating_flow_temperature_from_heat_distribution_system,
                    summer_heating_mode=summer_heating_mode,
                    storage_temperature_modifier=storage_temperature_modifier,
                )

            # mode 2 is regulated controller (meaning heating, cooling, off). this is only possible if heating system is floor heating
            elif self.mode == 2 and self.heat_distribution_system_type == HeatDistributionSystemType.FLOORHEATING:
                # turning heat pump cooling mode off when the average daily outside temperature is below a certain threshold
                summer_cooling_mode = self.summer_cooling_condition(
                    daily_average_outside_temperature_in_celsius=daily_avg_outside_temperature_in_celsius,
                    set_cooling_threshold_temperature_in_celsius=self.heatpump_controller_config.set_cooling_threshold_outside_temperature_in_celsius,
                )
                self.conditions_heating_cooling_off(
                    water_temperature_input_in_celsius=water_temperature_input_from_heat_water_storage_in_celsius,
                    set_heating_flow_temperature_in_celsius=heating_flow_temperature_from_heat_distribution_system,
                    summer_heating_mode=summer_heating_mode,
                    summer_cooling_mode=summer_cooling_mode,
                    storage_temperature_modifier=storage_temperature_modifier,
                )

            else:
                raise ValueError(
                    "Either the Advanced HP Lib Controller Mode is neither 1 nor 2,"
                    "or the heating system is not floor heating which is the condition for cooling (mode 2)."
                )

            if self.controller_heatpumpmode == "heating":
                state = 1
            elif self.controller_heatpumpmode == "cooling":
                state = -1
            elif self.controller_heatpumpmode == "off":
                state = 0
            else:
                raise ValueError("Advanced HP Lib Controller State unknown.")

            stsv.set_output_value(self.state_channel, state)

    def conditions_on_off(
        self,
        water_temperature_input_in_celsius: float,
        set_heating_flow_temperature_in_celsius: float,
        summer_heating_mode: str,
        storage_temperature_modifier: float,
    ) -> None:
        """Set conditions for the heat pump controller mode."""

        if self.controller_heatpumpmode == "heating":
            if (
                water_temperature_input_in_celsius
                > (set_heating_flow_temperature_in_celsius + storage_temperature_modifier)
                or summer_heating_mode == "off"
            ):
                self.controller_heatpumpmode = "off"
                return

        elif self.controller_heatpumpmode == "off":
            # heat pump is only turned on if the water temperature is below the flow temperature
            # and if the avg daily outside temperature is cold enough (summer mode on)
            if (
                water_temperature_input_in_celsius
                < (set_heating_flow_temperature_in_celsius + storage_temperature_modifier)
                and summer_heating_mode == "on"
            ):
                self.controller_heatpumpmode = "heating"
                return

        else:
            raise ValueError("unknown mode")

    def conditions_heating_cooling_off(
        self,
        water_temperature_input_in_celsius: float,
        set_heating_flow_temperature_in_celsius: float,
        summer_heating_mode: str,
        summer_cooling_mode: str,
        storage_temperature_modifier: float,
    ) -> None:
        """Set conditions for the heat pump controller mode according to the flow temperature."""
        # Todo: storage temperature modifier is only working for heating so far. Implement for cooling similar
        heating_set_temperature = set_heating_flow_temperature_in_celsius
        cooling_set_temperature = set_heating_flow_temperature_in_celsius

        if self.controller_heatpumpmode == "heating":
            if (
                water_temperature_input_in_celsius
                >= heating_set_temperature
                + storage_temperature_modifier  # Todo: Check if storage_temperature_modifier is neccessary here
                or summer_heating_mode == "off"
            ):
                self.controller_heatpumpmode = "off"
                return
        elif self.controller_heatpumpmode == "cooling":
            if water_temperature_input_in_celsius <= cooling_set_temperature or summer_cooling_mode == "off":
                self.controller_heatpumpmode = "off"
                return

        elif self.controller_heatpumpmode == "off":
            # heat pump is only turned on if the water temperature is below the flow temperature
            # and if the avg daily outside temperature is cold enough (summer heating mode on)
            if (
                water_temperature_input_in_celsius < (heating_set_temperature + storage_temperature_modifier)
                and summer_heating_mode == "on"
            ):
                self.controller_heatpumpmode = "heating"
                return

            # heat pump is only turned on for cooling if the water temperature is above a certain flow temperature
            # and if the avg daily outside temperature is warm enough (summer cooling mode on)
            if water_temperature_input_in_celsius > (cooling_set_temperature) and summer_cooling_mode == "on":
                self.controller_heatpumpmode = "cooling"
                return

        else:
            raise ValueError("unknown mode")

    def summer_heating_condition(
        self,
        daily_average_outside_temperature_in_celsius: float,
        set_heating_threshold_temperature_in_celsius: Optional[float],
    ) -> str:
        """Set conditions for the heat pump."""

        # if no heating threshold is set, the heat pump is always on
        if set_heating_threshold_temperature_in_celsius is None:
            heating_mode = "on"

        # it is too hot for heating
        elif daily_average_outside_temperature_in_celsius > set_heating_threshold_temperature_in_celsius:
            heating_mode = "off"

        # it is cold enough for heating
        elif daily_average_outside_temperature_in_celsius < set_heating_threshold_temperature_in_celsius:
            heating_mode = "on"

        else:
            raise ValueError(
                f"daily average temperature {daily_average_outside_temperature_in_celsius}°C"
                f"or heating threshold temperature {set_heating_threshold_temperature_in_celsius}°C is not acceptable."
            )
        return heating_mode

    def summer_cooling_condition(
        self,
        daily_average_outside_temperature_in_celsius: float,
        set_cooling_threshold_temperature_in_celsius: Optional[float],
    ) -> str:
        """Set conditions for the heat pump."""

        # if no cooling threshold is set, cooling is always possible no matter what daily outside temperature
        if set_cooling_threshold_temperature_in_celsius is None:
            cooling_mode = "on"

        # it is hot enough for cooling
        elif daily_average_outside_temperature_in_celsius > set_cooling_threshold_temperature_in_celsius:
            cooling_mode = "on"

        # it is too cold for cooling
        elif daily_average_outside_temperature_in_celsius < set_cooling_threshold_temperature_in_celsius:
            cooling_mode = "off"

        else:
            raise ValueError(
                f"daily average temperature {daily_average_outside_temperature_in_celsius}°C"
                f"or cooling threshold temperature {set_cooling_threshold_temperature_in_celsius}°C is not acceptable."
            )

        return cooling_mode

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of electricity costs and revenues."""
        opex_cost_data_class = OpexCostDataClass.get_default_opex_cost_data_class()
        return opex_cost_data_class

    @staticmethod
    def get_cost_capex(config: HeatPumpHplibControllerL1Config, simulation_parameters: SimulationParameters) -> CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []


@dataclass
class CalculationRequest(JSONWizard):
    """Class for caching hplib parameters so that hplib.simulate does not need to run so often."""

    t_in_primary: float
    t_in_secondary: float
    t_amb: float
    mode: int

    def get_key(self):
        """Get key of class with important parameters."""

        return str(self.t_in_primary) + " " + str(self.t_in_secondary) + " " + str(self.t_amb) + " " + str(self.mode)



"""Result Path Provider Module."""

# clean
import sys
import os
import datetime
import enum
from typing import Any, Optional

from hisim.sim_repository_singleton import SingletonMeta


class ResultPathProviderSingleton(metaclass=SingletonMeta):

    """ResultPathProviderSingleton class.

    According to your storting options and your input information a result path is created.
    """

    def __init__(
        self,
    ):
        """Initialize the class."""
        self.base_path: Optional[str] = None
        self.model_name: Optional[str] = None
        self.variant_name: Optional[str] = None
        self.scenario_hash_string: Optional[str] = None
        self.further_result_folder_description: Optional[str] = None
        self.sorting_option: Any = SortingOptionEnum.FLAT
        self.time_resolution_in_seconds: Optional[int] = None
        self.simulation_duration_in_days: Optional[int] = None
        self.datetime_string: str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    def set_important_result_path_information(
        self,
        module_directory: str,
        model_name: str,
        variant_name: Optional[str],
        scenario_hash_string: Optional[str],
        sorting_option: Any,
        further_result_folder_description: Optional[str] = None,
    ) -> None:
        """Set important result path information."""
        self.set_base_path(module_directory=module_directory)
        self.set_model_name(model_name=model_name)
        self.set_variant_name(variant_name=variant_name)
        self.set_sorting_option(sorting_option=sorting_option)
        self.set_scenario_hash_string(scenario_hash_string=scenario_hash_string)
        self.set_further_result_folder_description(further_result_folder_description=further_result_folder_description)

    def set_base_path(self, module_directory: str) -> None:
        """Set base path."""
        self.base_path = os.path.join(module_directory, "results")

    def set_model_name(self, model_name: str) -> None:
        """Set model name."""
        self.model_name = model_name

    def set_variant_name(self, variant_name: Optional[str]) -> None:
        """Set variant name."""
        if variant_name is None:
            variant_name = ""
        self.variant_name = variant_name

    def set_scenario_hash_string(self, scenario_hash_string: Optional[str]) -> None:
        """Set scenario hash string."""
        if scenario_hash_string is None:
            scenario_hash_string = ""

        self.scenario_hash_string = scenario_hash_string

    def set_further_result_folder_description(self, further_result_folder_description: Optional[str]) -> None:
        """Set further result folder description."""
        self.further_result_folder_description = further_result_folder_description

    def set_sorting_option(self, sorting_option: Any) -> None:
        """Set sorting option."""
        self.sorting_option = sorting_option

    def set_time_resolution(self, time_resolution_in_seconds: int) -> None:
        """Set time resolution."""
        self.time_resolution_in_seconds = time_resolution_in_seconds

    def set_simulation_duration(self, simulation_duration_in_days: int) -> None:
        """Set simulation duration."""
        self.simulation_duration_in_days = simulation_duration_in_days

    def get_result_directory_name(self) -> Any:
        """Get the result directory path."""

        if (
            self.base_path is not None
            and self.model_name is not None
            and self.variant_name is not None
            and self.datetime_string is not None
            and self.scenario_hash_string is not None
        ):
            if self.sorting_option == SortingOptionEnum.DEEP:
                path = os.path.join(
                    self.base_path,
                    self.model_name,
                    self.variant_name,
                    self.datetime_string,
                )
            elif self.sorting_option == SortingOptionEnum.MASS_SIMULATION_WITH_INDEX_ENUMERATION:
                # schauen ob verzeichnis schon da und aufsteigende nummer anhängen
                idx = 1

                if self.further_result_folder_description is not None:
                    path = os.path.join(
                        self.base_path,
                        self.model_name,
                        self.further_result_folder_description,
                        self.variant_name + "_" + str(idx),
                    )
                    while os.path.exists(path):
                        idx = idx + 1
                        path = os.path.join(
                            self.base_path,
                            self.model_name,
                            self.further_result_folder_description,
                            self.variant_name + "_" + str(idx),
                        )
                else:
                    path = os.path.join(
                        self.base_path,
                        self.model_name,
                        self.variant_name + "_" + str(idx),
                    )
                    while os.path.exists(path):
                        idx = idx + 1
                        path = os.path.join(
                            self.base_path,
                            self.model_name,
                            self.variant_name + "_" + str(idx),
                        )
            elif self.sorting_option == SortingOptionEnum.MASS_SIMULATION_WITH_HASH_ENUMERATION:
                if self.further_result_folder_description is not None:
                    path = os.path.join(
                        self.base_path,
                        self.model_name,
                        self.further_result_folder_description,
                        self.variant_name + "_" + self.scenario_hash_string,
                    )
                else:
                    path = os.path.join(
                        self.base_path,
                        self.model_name,
                        self.variant_name + "_" + self.scenario_hash_string,
                    )

            elif self.sorting_option == SortingOptionEnum.FLAT:
                path = os.path.join(
                    self.base_path,
                    self.model_name + "_" + self.variant_name + self.datetime_string,
                )

            check_path_length(path=path)
            return path

        return None


def check_path_length(path: str) -> None:
    """Make sure that path name does not get too long for Windows."""

    character_limit_according_to_windows = 256
    # check if the system is windows
    is_windows = sys.platform.startswith("win")
    if is_windows and len(path) >= character_limit_according_to_windows:
        raise NameError(
            f"The path {path} exceeds the limit of 256 characters which is the limit for Windows. Please make your path shorter."
        )


class SortingOptionEnum(enum.Enum):

    """A SortingOptionEnum class."""

    DEEP = 1
    MASS_SIMULATION_WITH_INDEX_ENUMERATION = 2
    MASS_SIMULATION_WITH_HASH_ENUMERATION = 3
    FLAT = 4



"""Implementation of shiftable household devices like washing machines, dish washers or dryers.

Takes load profiles and time windows, where the activation can be shifted within from LoadProfileGenerator and activates the device when surplus from PV is available.
The device is activated at the end of the time window when no surplus was available. This file contains the class SmartDevice and SmartDevice State,
the configuration is automatically adopted from the information provided by the LPG.
"""

# clean

# Generic/Built-in
import json
import math as ma
from os import path
from typing import List, Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import pandas as pd


# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters
from hisim.component import OpexCostDataClass
from hisim.postprocessing.kpi_computation.kpi_structure import KpiTagEnumClass

__authors__ = "Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class SmartDeviceConfig(cp.ConfigBase):
    """Configuration of the smart device."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return SmartDevice.get_full_classname()

    building_name: str
    name: str
    identifier: str
    source_weight: int
    smart_devices_included: bool

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config."""
        return SmartDeviceConfig(
            building_name=building_name,
            name="Smart Device",
            identifier="Identifier",
            source_weight=1,
            smart_devices_included=True,
        )


class SmartDeviceState:
    """State representing smart appliance."""

    def __init__(
        self,
        actual_power: float = 0,
        timestep_of_activation: int = -999,
        time_to_go: int = 0,
        position: int = 0,
    ):
        """Initilization of state.

        :param actual_power: power of smart appliance at given timestep, defaults to 0
        :type actual_power: float, optional
        :param timestep_of_activation: timestep, where the device was activated, defaults to -999
        :type timestep_of_activation: int, optional
        :param time_to_go: duration of the power profile, which follows for the nex time steps, defaults to 0
        :type time_to_go: int, optional
        :param position: index of demand profile relevent for the given timestep, defaults to 0
        :type position: int, optional
        """
        self.actual_power = actual_power
        self.timestep_of_activation = timestep_of_activation
        self.time_to_go = time_to_go
        self.position = position

    def clone(self) -> "SmartDeviceState":
        """Copy state efficiently."""
        return SmartDeviceState(
            self.actual_power,
            self.timestep_of_activation,
            self.time_to_go,
            self.position,
        )

    def run(self, timestep: int, electricity_profile: List[float]) -> None:
        """Check device state based on previous time step.

        :param timestep: timestep of simulation
        :type timestep: int
        :param electricity_profile: load profile of device for actual or next activation
        :type electricity_profile: List[float]
        """
        # device activation
        if timestep > self.timestep_of_activation + self.time_to_go:
            self.timestep_of_activation = timestep
            self.time_to_go = len(electricity_profile)
            self.actual_power = electricity_profile[0]

        if timestep < self.timestep_of_activation + self.time_to_go:
            # device is running
            self.actual_power = electricity_profile[timestep - self.timestep_of_activation]

        # device deactivation
        if timestep == self.timestep_of_activation + self.time_to_go:
            self.position = self.position + 1
            self.time_to_go = 0
            self.actual_power = 0


class SmartDevice(cp.Component):
    """Smart device class.

    Class component that provides availablity and profiles of flexible smart devices like shiftable (in time) washing machines and dishwashers.
    Data provided or based on LPG exports.
    """

    # mandatory Inputs
    L3DeviceActivation = "l3_DeviceActivation"

    # mandatory Outputs
    ElectricityOutput = "ElectricityOutput"

    # optional Inputs
    ElectricityTarget = "ElectricityTarget"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: SmartDeviceConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ):
        """Initialize the class."""

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.build(
            identifier=config.identifier,
            source_weight=config.source_weight,
            seconds_per_timestep=my_simulation_parameters.seconds_per_timestep,
        )
        self.previous_state: SmartDeviceState
        self.state: SmartDeviceState
        self.consumption = 0
        if my_simulation_parameters.surplus_control and config.smart_devices_included:
            postprocessing_flag = [
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
                lt.ComponentType.SMART_DEVICE,
            ]
        else:
            postprocessing_flag = [lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED]

        # mandatory Output
        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=postprocessing_flag,
            output_description="Electricity output",
        )

        self.electricity_target_channel: cp.ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.ElectricityTarget,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            mandatory=False,
        )

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Iteration in smart appliance like washing mashine, dish washer or dryer.

        :param timestep: timestep of simulation
        :type timestep: int
        :param stsv: _description_
        :type stsv: cp.SingleTimeStepValues
        :param force_convergence: _description_
        :type force_convergence: bool
        """

        # initialize power
        self.state.actual_power = 0

        # if not already running: check if activation makes sense
        if timestep > self.state.timestep_of_activation + self.state.time_to_go:
            if timestep > self.earliest_start[self.state.position]:  # can be turnod on
                # initialize next activation
                activation: float = timestep + 10
                # if surplus controller is connected get related signal
                if self.electricity_target_channel.source_output is not None:
                    electricity_target = stsv.get_input_value(self.electricity_target_channel)
                    if electricity_target >= self.electricity_profile[self.state.position][0]:
                        activation = timestep
                # if last possible switch on force activation
                if timestep >= self.latest_start[self.state.position]:  # needs to be activated
                    activation = timestep

                if timestep == activation:
                    self.state.run(timestep, self.electricity_profile[self.state.position])

        # run device if it was already activated
        else:
            self.state.run(timestep, self.electricity_profile[self.state.position])

        stsv.set_output_value(self.electricity_output_channel, self.state.actual_power)

    def build(self, identifier: str, source_weight: int, seconds_per_timestep: int = 60) -> None:
        """Initialization of Smart Device information.

        :param identifier: name of smart device in LPG
        :type identifier: str
        :param source_weight: priority of smart device in Energy Management System
        :type source_weight: int
        :param seconds_per_timestep: time step size, defaults to 60
        :type seconds_per_timestep: int, optional
        :raises NameError: _description_
        :raises TypeError: _description_
        """

        # load smart device profile
        smart_device_profile = []
        filepath = path.join(utils.HISIMPATH["utsp_reports"], "FlexibilityEvents.HH1.json")
        with open(filepath, encoding="utf-8") as file:
            smart_device_profile = json.load(file)

        if not smart_device_profile:
            raise NameError("LPG data for smart appliances is missing or located missleadingly")

        # initializing relevant data
        earliest_start, latest_start, electricity_profile = [], [], []

        minutes_per_timestep = seconds_per_timestep / 60

        if not minutes_per_timestep.is_integer():
            raise TypeError(
                "Up to now smart appliances have only been implemented for time resolutions corresponding to multiples of one minute"
            )
        minutes_per_timestep = int(minutes_per_timestep)

        # reading in data from json file and adopting to given time resolution
        for sample in smart_device_profile:
            device_name = str(sample["Device"]["Name"])
            if device_name == identifier:
                # earliest start in given time resolution -> integer value
                x_sample = sample["EarliestStart"]["ExternalStep"]
                # skip if occurs in calibration days (negative sign )
                if x_sample < 0:
                    continue
                # timestep (in minutes) the profile is shifted in the first step of the external time resolution
                offset = minutes_per_timestep - x_sample % minutes_per_timestep
                # earliest start in given time resolution -> float value
                x_sample = x_sample / minutes_per_timestep
                # latest start in given time resolution
                y_sample = sample["LatestStart"]["ExternalStep"] / minutes_per_timestep
                # number of timesteps in given time resolution -> integer value
                z_sample = ma.ceil(x_sample + sample["TotalDuration"] / minutes_per_timestep) - ma.floor(x_sample)
                # earliest and latest start in new time resolution -> integer value
                earliest_start.append(ma.floor(x_sample))
                latest_start.append(ma.ceil(y_sample))

                # get shiftable load profile
                el_shiftable_load = sample["Profiles"][2]["TimeOffsetInSteps"] * [0] + sample["Profiles"][2]["Values"]

                # average profiles given in 1 minute resolution to given time resolution
                elem_el = []
                # append first timestep which may not fill  the entire 15 minutes
                elem_el.append(sum(el_shiftable_load[:offset]) / offset)

                i = 0
                for i in range(z_sample - 2):
                    elem_el.append(
                        sum(
                            el_shiftable_load[
                                offset + minutes_per_timestep * i : offset + (i + 1) * minutes_per_timestep
                            ]
                        )
                        / minutes_per_timestep
                    )

                last = el_shiftable_load[offset + (i + 1) * minutes_per_timestep :]
                if offset != minutes_per_timestep:
                    elem_el.append(sum(last) / (minutes_per_timestep - offset))
                electricity_profile.append(elem_el)

        self.source_weight = source_weight
        earliest_start = earliest_start + [
            self.my_simulation_parameters.timesteps
        ]  # append value to continue simulation after last necesary run of flexible device at end of year
        self.earliest_start = utils.convert_lpg_timestep_to_utc(
            data=earliest_start,
            year=self.my_simulation_parameters.year,
            seconds_per_timestep=seconds_per_timestep,
        )
        latest_start = latest_start + [
            self.my_simulation_parameters.timesteps + 999
        ]  # append value to continue simulation after last necesary run of smart device at end of year
        self.latest_start = utils.convert_lpg_timestep_to_utc(
            data=latest_start,
            year=self.my_simulation_parameters.year,
            seconds_per_timestep=seconds_per_timestep,
        )
        self.electricity_profile = electricity_profile
        self.state = SmartDeviceState()
        self.previous_state = SmartDeviceState()

    def write_to_report(self) -> List[str]:
        """Writes relevant information to report."""
        lines: List[str] = []
        lines.append(f"DeviceName: {self.component_name}")
        lines.append(f"Consumption: {self.consumption:.2f}")
        return lines

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Get opex costs."""
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name and output.load_type == lt.LoadTypes.ELECTRICITY:
                self.consumption = (
                    sum(postprocessing_results.iloc[:, index])
                    * self.my_simulation_parameters.seconds_per_timestep
                    / 3.6e6
                )
        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=0,  # TODO: add maintenance costs
            co2_footprint_in_kg=0,
            consumption_in_kwh=self.consumption,
            loadtype=lt.LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.SMART_DEVICE
        )

        return opex_cost_data_class
